# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Here are general and special bokeh plots to use

"""
from .bokeh_plotter import BokehPlotter
from .parameters import ensure_plotter_consistency, NestedPlotParameters
from .data import process_data_arguments

import pandas as pd
import numpy as np
import warnings
from pprint import pprint

################## Helpers     ################

plot_params = BokehPlotter()


def set_bokeh_plot_defaults(**kwargs):
    """
    Set defaults for bokeh backend
    according to the given keyword arguments

    Available defaults can be seen in :py:class:`~masci_tools.vis.bokeh_plotter.BokehPlotter`
    """
    plot_params.set_defaults(**kwargs)


def reset_bokeh_plot_defaults():
    """
    Reset the defaults for bokeh backend
    to the hardcoded defaults

    Available defaults can be seen in :py:class:`~masci_tools.vis.bokeh_plotter.BokehPlotter`
    """
    plot_params.reset_defaults()


def show_bokeh_plot_defaults():
    """
    Show the currently set defaults for bokeh backend

    Available defaults can be seen in :py:class:`~masci_tools.vis.bokeh_plotter.BokehPlotter`
    """
    pprint(plot_params.get_dict())


def get_bokeh_help(key):
    """
    Print the decription of the given key in the bokeh backend

    Available defaults can be seen in :py:class:`~masci_tools.vis.bokeh_plotter.BokehPlotter`
    """
    plot_params.get_description(key)


def load_bokeh_defaults(filename='plot_bokeh_defaults.json'):
    """
    Load defaults for the bokeh backend from a json file.

    :param filename: filename,from  where the defaults should be taken
    """
    plot_params.load_defaults(filename)


def save_bokeh_defaults(filename='plot_bokeh_defaults.json', save_complete=False):
    """
    Save the current defaults for the matplotlib backend to a json file.

    :param filename: filename, where the defaults should be stored
    :param save_complete: bool if True not only the overwritten user defaults
                          but also the unmodified harcoded defaults are stored
    """
    plot_params.save_defaults(filename, save_complete=save_complete)


##################################### general plots ##########################


@ensure_plotter_consistency(plot_params)
def bokeh_scatter(x,
                  y=None,
                  *,
                  xlabel='x',
                  ylabel='y',
                  title='',
                  figure=None,
                  data=None,
                  saveas='scatter',
                  copy_data=False,
                  **kwargs):
    """
    Create an interactive scatter plot with bokeh

    :param x: arraylike or key for data for the x-axis
    :param y: arraylike or key for data for the y-axis
    :param data: source for the data of the plot (pandas Dataframe for example)
    :param xlabel: label for the x-axis
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param figure: bokeh figure (optional), if provided the plot will be added to this figure
    :param outfilename: filename of the output file
    :param copy_data: bool, if True the data argument will be copied

    Kwargs will be passed on to :py:class:`masci_tools.vis.bokeh_plotter.BokehPlotter`.
    If the arguments are not recognized they are passed on to the bokeh function `scatter`
    """
    from bokeh.models import ColumnDataSource

    if isinstance(x, (dict, pd.DataFrame, ColumnDataSource)) or x is None:
        warnings.warn(
            'Passing the source as first argument is deprecated. Please pass in source by the keyword data'
            'and xdata and ydata as the first arguments', DeprecationWarning)
        data = x
        x = kwargs.pop('xdata', 'x')
        y = kwargs.pop('ydata', 'y')

    plot_data = process_data_arguments(data=data,
                                       x=x,
                                       y=y,
                                       copy_data=copy_data,
                                       single_plot=True,
                                       same_length=True,
                                       use_column_source=True)
    entry, source = plot_data.items(first=True)

    plot_params.set_defaults(default_type='function', name=entry.y)
    kwargs = plot_params.set_parameters(continue_on_error=True, **kwargs)

    p = plot_params.prepare_figure(title, xlabel, ylabel, figure=figure)

    plot_kwargs = plot_params.plot_kwargs(plot_type='scatter')
    res = p.scatter(x=entry.x, y=entry.y, source=source, **plot_kwargs, **kwargs)
    plot_params.add_tooltips(p, res, entry)

    if plot_params['level'] is not None:
        res.level = plot_params['level']

    plot_params.draw_straight_lines(p)
    plot_params.set_limits(p)
    plot_params.save_plot(p, saveas)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_multi_scatter(x,
                        y=None,
                        *,
                        data=None,
                        figure=None,
                        xlabel='x',
                        ylabel='y',
                        title='',
                        saveas='scatter',
                        copy_data=False,
                        set_default_legend=True,
                        **kwargs):
    """
    Create an interactive scatter (muliple data sets possible) plot with bokeh

    :param x: arraylike or key for data for the x-axis
    :param y: arraylike or key for data for the y-axis
    :param data: source for the data of the plot (pandas Dataframe for example)
    :param xlabel: label for the x-axis
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param figure: bokeh figure (optional), if provided the plot will be added to this figure
    :param outfilename: filename of the output file
    :param copy_data: bool, if True the data argument will be copied
    :param set_default_legend: bool if True the data names are used to generate default legend labels

    Kwargs will be passed on to :py:class:`masci_tools.vis.bokeh_plotter.BokehPlotter`.
    If the arguments are not recognized they are passed on to the bokeh function `scatter`
    """
    from bokeh.models import ColumnDataSource

    if isinstance(x, (dict, pd.DataFrame, ColumnDataSource)) or x is None:
        warnings.warn(
            'Passing the source as first argument is deprecated. Please pass in source by the keyword data'
            'and xdata and ydata as the first arguments', DeprecationWarning)
        data = x
        x = kwargs.pop('xdata', 'x')
        y = kwargs.pop('ydata', 'y')

    plot_data = process_data_arguments(data=data,
                                       x=x,
                                       y=y,
                                       same_length=True,
                                       copy_data=copy_data,
                                       use_column_source=True)

    plot_params.single_plot = False
    plot_params.num_plots = len(plot_data)

    if plot_data.distinct_datasets('x') == 1:
        default_legend_label = plot_data.get_keys('y')
    else:
        default_legend_label = plot_data.get_keys('x')
    if set_default_legend:
        plot_params.set_defaults(default_type='function', legend_label=default_legend_label)

    plot_params.set_defaults(default_type='function', name=default_legend_label)
    kwargs = plot_params.set_parameters(continue_on_error=True, **kwargs)
    p = plot_params.prepare_figure(title, xlabel, ylabel, figure=figure)

    #Process the given color arguments
    plot_params.set_color_palette_by_num_plots()

    plot_kwargs = plot_params.plot_kwargs(plot_type='scatter')

    for indx, ((entry, source), plot_kw) in enumerate(zip(plot_data.items(), plot_kwargs)):

        res = p.scatter(x=entry.x, y=entry.y, source=source, **plot_kw, **kwargs)
        plot_params.add_tooltips(p, res, entry)

        if plot_params[('level', indx)] is not None:
            res.level = plot_params[('level', indx)]

    plot_params.draw_straight_lines(p)
    plot_params.set_limits(p)
    plot_params.set_legend(p)
    plot_params.save_plot(p, saveas)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_line(x,
               y=None,
               *,
               data=None,
               figure=None,
               xlabel='x',
               ylabel='y',
               title='',
               saveas='line',
               plot_points=False,
               area_curve=0,
               copy_data=False,
               set_default_legend=True,
               **kwargs):
    """
    Create an interactive multi-line plot with bokeh

    :param x: arraylike or key for data for the x-axis
    :param y: arraylike or key for data for the y-axis
    :param data: source for the data of the plot (optional) (pandas Dataframe for example)
    :param xlabel: label for the x-axis
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param figure: bokeh figure (optional), if provided the plot will be added to this figure
    :param outfilename: filename of the output file
    :param plot_points: bool, if True also plot the points with a scatterplot on top
    :param copy_data: bool, if True the data argument will be copied
    :param set_default_legend: bool if True the data names are used to generate default legend labels

    Kwargs will be passed on to :py:class:`masci_tools.vis.bokeh_plotter.BokehPlotter`.
    If the arguments are not recognized they are passed on to the bokeh function `line`
    """
    from bokeh.models import ColumnDataSource

    if isinstance(x, (dict, pd.DataFrame, ColumnDataSource)) or x is None:
        warnings.warn(
            'Passing the source as first argument is deprecated. Please pass in source by the keyword data'
            'and xdata and ydata as the first arguments', DeprecationWarning)
        data = x
        x = kwargs.pop('xdata', 'x')
        y = kwargs.pop('ydata', 'y')

    plot_data = process_data_arguments(data=data,
                                       x=x,
                                       y=y,
                                       shift=area_curve,
                                       same_length=True,
                                       copy_data=copy_data,
                                       use_column_source=True)

    plot_params.single_plot = False
    plot_params.num_plots = len(plot_data)

    if plot_data.distinct_datasets('x') == 1:
        default_legend_label = plot_data.get_keys('y')
    else:
        default_legend_label = plot_data.get_keys('x')
    if set_default_legend:
        plot_params.set_defaults(default_type='function', legend_label=default_legend_label)

    plot_params.set_defaults(default_type='function', name=default_legend_label)

    kwargs = plot_params.set_parameters(continue_on_error=True, **kwargs)
    p = plot_params.prepare_figure(title, xlabel, ylabel, figure=figure)

    #Process the given color arguments
    plot_params.set_color_palette_by_num_plots()

    plot_kw_line = plot_params.plot_kwargs(plot_type='line')
    plot_kw_scatter = plot_params.plot_kwargs(plot_type='scatter')
    plot_kw_area = plot_params.plot_kwargs(plot_type='area')

    area_curve = kwargs.pop('area_curve', None)

    for indx, ((entry, source), kw_line, kw_scatter,
               kw_area) in enumerate(zip(plot_data.items(), plot_kw_line, plot_kw_scatter, plot_kw_area)):

        if plot_params[('area_plot', indx)]:
            if plot_params[('area_vertical', indx)]:
                p.harea(y=entry.y, x1=entry.x, x2=entry.shift, **kw_area, source=source)
            else:
                p.varea(x=entry.x, y1=entry.y, y2=entry.shift, **kw_area, source=source)

        res = p.line(x=entry.x, y=entry.y, source=source, **kw_line, **kwargs)
        plot_params.add_tooltips(p, res, entry)
        res2 = None
        if plot_points:
            res2 = p.scatter(x=entry.x, y=entry.y, source=source, **kw_scatter)

        if plot_params[('level', indx)] is not None:
            res.level = plot_params[('level', indx)]
            if res2 is not None:
                res2.level = plot_params[('level', indx)]

    plot_params.draw_straight_lines(p)
    plot_params.set_limits(p)
    plot_params.set_legend(p)
    plot_params.save_plot(p, saveas)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_dos(energy_grid,
              dos_data=None,
              *,
              data=None,
              energy_label=r'E-E_F [eV]',
              dos_label=r'DOS [1/eV]',
              title=r'Density of states',
              xyswitch=False,
              e_fermi=0,
              saveas='dos_plot',
              copy_data=False,
              **kwargs):
    """
    Create an interactive dos plot (non-spinpolarized) with bokeh
    Both horizontal or vertical orientation are possible

    :param energy_grid: arraylike or key data for the energy grid
    :param spin_up_data: arraylike or key data for the DOS
    :param data: source for the DOS data (optional) of the plot (pandas Dataframe for example)
    :param energy_label: label for the energy-axis
    :param dos_label: label for the dos-axis
    :param title: title of the figure
    :param xyswitch: bool if True, the energy will be plotted along the y-direction
    :param e_fermi: float, determines, where to put the line for the fermi energy
    :param outfilename: filename of the output file
    :param copy_data: bool, if True the data argument will be copied

    Kwargs will be passed on to :py:func:`bokeh_line()`
    """
    from bokeh.models import ColumnDataSource

    if isinstance(energy_grid, (dict, pd.DataFrame, ColumnDataSource)) or energy_grid is None:
        warnings.warn(
            'Passing the dataframe as first argument is deprecated. Please pass in source by the keyword data'
            'and energy_grid and dos_data as the first arguments', DeprecationWarning)
        data = energy_grid
        energy_grid = kwargs.pop('energy', 'energy_grid')
        dos_data = kwargs.pop('ynames', None)

    if dos_data is None and data is not None:
        dos_data = set(data.keys()) - set([energy_grid] if isinstance(energy_grid, str) else energy_grid)
        dos_data = sorted(dos_data)

    plot_data = process_data_arguments(data=data,
                                       energy=energy_grid,
                                       dos=dos_data,
                                       same_length=True,
                                       copy_data=copy_data,
                                       use_column_source=True)

    plot_params.single_plot = False
    plot_params.num_plots = len(plot_data)

    if 'limits' in kwargs:
        limits = kwargs.pop('limits')
        if 'x' not in limits and 'y' not in limits:
            if xyswitch:
                limits['x'], limits['y'] = limits.pop('dos', None), limits.pop('energy', None)
            else:
                limits['x'], limits['y'] = limits.pop('energy', None), limits.pop('dos', None)
        kwargs['limits'] = {k: v for k, v in limits.items() if v is not None}

    lines = {'horizontal': 0}
    lines['vertical'] = e_fermi

    if xyswitch:
        lines['vertical'], lines['horizontal'] = lines['horizontal'], lines['vertical']

    plot_params.set_defaults(default_type='function',
                             straight_lines=lines,
                             tooltips=[('Name', '$name'), ('Energy', '@{x}{{0.0[00]}}'),
                                       ('DOS value', '@$name{{0.00}}')],
                             figure_kwargs={
                                 'width': 1000,
                             })

    if xyswitch:
        x, y = plot_data.get_keys('dos'), plot_data.get_keys('energy')
        xlabel, ylabel = dos_label, energy_label
        plot_params.set_defaults(default_type='function', area_vertical=True)
    else:
        xlabel, ylabel = energy_label, dos_label
        x, y = plot_data.get_keys('energy'), plot_data.get_keys('dos')

    p = bokeh_line(x,
                   y,
                   data=plot_data.data,
                   xlabel=xlabel,
                   ylabel=ylabel,
                   title=title,
                   name=y,
                   saveas=saveas,
                   **kwargs)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_spinpol_dos(energy_grid,
                      spin_up_data=None,
                      spin_dn_data=None,
                      *,
                      data=None,
                      spin_dn_negative=True,
                      energy_label=r'E-E_F [eV]',
                      dos_label=r'DOS [1/eV]',
                      title=r'Density of states',
                      xyswitch=False,
                      e_fermi=0,
                      spin_arrows=True,
                      saveas='dos_plot',
                      copy_data=False,
                      **kwargs):
    """
    Create an interactive dos plot (spinpolarized) with bokeh
    Both horizontal or vertical orientation are possible

    :param energy_grid: arraylike or key data for the energy grid
    :param spin_up_data: arraylike or key data for the DOS spin-up
    :param spin_dn_data: arraylike or key data for the DOS spin-dn
    :param data: source for the DOS data (optional) of the plot (pandas Dataframe for example)
    :param spin_dn_negative: bool, if True (default), the spin down components are plotted downwards
    :param energy_label: label for the energy-axis
    :param dos_label: label for the dos-axis
    :param title: title of the figure
    :param xyswitch: bool if True, the energy will be plotted along the y-direction
    :param e_fermi: float, determines, where to put the line for the fermi energy
    :param spin_arrows: bool, if True (default) small arrows will be plotted on the left side of the plot indicating
                        the spin directions (if spin_dn_negative is True)
    :param outfilename: filename of the output file
    :param copy_data: bool, if True the data argument will be copied

    Kwargs will be passed on to :py:func:`bokeh_line()`
    """
    from bokeh.models import NumeralTickFormatter, Arrow, NormalHead
    from bokeh.models import ColumnDataSource

    if isinstance(energy_grid, (dict, pd.DataFrame, ColumnDataSource)) or energy_grid is None:
        warnings.warn(
            'Passing the dataframe as first argument is deprecated. Please pass in source by the keyword data'
            'and energy_grid and dos_data as the first arguments', DeprecationWarning)
        data = energy_grid
        energy_grid = kwargs.pop('energy', 'energy_grid')
        spin_up_data = kwargs.pop('ynames', None)
        spin_up_data, spin_dn_data = spin_up_data[:len(spin_up_data) // 2], spin_up_data[len(spin_up_data) // 2:]

    if spin_up_data is None and data is not None:
        spin_up_data = set(key for key in data.keys() if '_up' in key)
        spin_up_data = sorted(spin_up_data)
        spin_dn_data = set(key for key in data.keys() if '_dn' in key)
        spin_dn_data = sorted(spin_dn_data)

    plot_data = process_data_arguments(data=data,
                                       energy=energy_grid,
                                       spin_up=spin_up_data,
                                       spin_dn=spin_dn_data,
                                       same_length=True,
                                       copy_data=copy_data,
                                       use_column_source=True)

    plot_params.single_plot = False
    plot_params.num_plots = len(plot_data)

    if 'limits' in kwargs:
        limits = kwargs.pop('limits')
        if 'x' not in limits and 'y' not in limits:
            if xyswitch:
                limits['x'], limits['y'] = limits.pop('dos', None), limits.pop('energy', None)
            else:
                limits['x'], limits['y'] = limits.pop('energy', None), limits.pop('dos', None)
        kwargs['limits'] = {k: v for k, v in limits.items() if v is not None}

    lines = {'horizontal': 0}
    lines['vertical'] = e_fermi

    if spin_dn_negative:
        plot_data.apply('spin_dn', lambda x: -x)

    if xyswitch:
        lines['vertical'], lines['horizontal'] = lines['horizontal'], lines['vertical']

    plot_params.set_defaults(default_type='function',
                             straight_lines=lines,
                             tooltips=[('DOS Name', '$name'), ('Energy', '@{x}{{0.0[00]}}'),
                                       ('Value', '@$name{{(0,0.00)}}')],
                             figure_kwargs={'width': 1000})

    #Create the full data for the scatterplot
    energy_entries = plot_data.get_keys('energy') * 2
    dos_entries = plot_data.get_keys('spin_up') + plot_data.get_keys('spin_dn')
    sources = plot_data.data
    if isinstance(sources, list):
        sources = sources * 2

    if xyswitch:
        x, y = dos_entries, energy_entries
        xlabel, ylabel = dos_label, energy_label
        plot_params.set_defaults(default_type='function',
                                 area_vertical=True,
                                 x_axis_formatter=NumeralTickFormatter(format='(0,0)'))
    else:
        xlabel, ylabel = energy_label, dos_label
        x, y = energy_entries, dos_entries
        plot_params.set_defaults(default_type='function',
                                 area_vertical=True,
                                 y_axis_formatter=NumeralTickFormatter(format='(0,0)'))

    plot_params.set_parameters(color=kwargs.pop('color', None), color_palette=kwargs.pop('color_palette', None))
    plot_params.set_color_palette_by_num_plots()

    #Double the colors for spin up and down
    kwargs['color'] = list(plot_params['color'].copy())
    kwargs['color'].extend(kwargs['color'])

    if 'legend_label' not in kwargs:
        kwargs['legend_label'] = dos_entries
    else:
        if isinstance(kwargs['legend_label'], list):
            if len(kwargs['legend_label']) == len(plot_data):
                kwargs['legend_label'].extend(kwargs['legend_label'])

    if 'show' in kwargs:
        plot_params.set_parameters(show=kwargs.pop('show'))
    if 'save_plots' in kwargs:
        plot_params.set_parameters(show=kwargs.pop('save_plots'))

    with NestedPlotParameters(plot_params):
        p = bokeh_line(x,
                       y,
                       xlabel=xlabel,
                       ylabel=ylabel,
                       title=title,
                       data=sources,
                       name=dos_entries,
                       show=False,
                       save_plots=False,
                       **kwargs)

    if spin_arrows and spin_dn_negative:

        #These are hardcoded because the parameters are not
        #reused anywhere (for now)
        x_pos = 50
        length = 70
        pad = 30
        height = p.plot_height - 100
        alpha = 0.5

        p.add_layout(
            Arrow(x_start=x_pos,
                  x_end=x_pos,
                  y_start=height - pad - length,
                  y_end=height - pad,
                  start_units='screen',
                  end_units='screen',
                  line_width=2,
                  line_alpha=alpha,
                  end=NormalHead(line_width=2, size=10, fill_alpha=alpha, line_alpha=alpha)))
        p.add_layout(
            Arrow(x_start=x_pos,
                  x_end=x_pos,
                  y_start=pad + length,
                  y_end=pad,
                  start_units='screen',
                  end_units='screen',
                  line_width=2,
                  line_alpha=alpha,
                  end=NormalHead(line_width=2, size=10, fill_alpha=alpha, line_alpha=alpha)))

    plot_params.save_plot(p, saveas)

    return p


@ensure_plotter_consistency(plot_params)
def bokeh_bands(kpath,
                bands=None,
                *,
                data=None,
                size_data=None,
                color_data=None,
                xlabel='',
                ylabel=r'E-E_F [eV]',
                title='',
                special_kpoints=None,
                markersize_min=3.0,
                markersize_scaling=10.0,
                saveas='bands_plot',
                scale_color=True,
                separate_bands=False,
                line_plot=False,
                band_index=None,
                copy_data=False,
                **kwargs):
    """
    Create an interactive bandstructure plot (non-spinpolarized) with bokeh
    Can make a simple plot or weight the size and color of the points against a given weight

    :param kpath: arraylike or key data for the kpoint data
    :param bands: arraylike or key data for the eigenvalues
    :param size_data: arraylike or key data the weights to emphasize (optional)
    :param color_data: str or arraylike, data for the color values with a colormap (optional)
    :param data: source for the bands data (optional) of the plot (pandas Dataframe for example)
    :param xlabel: label for the x-axis (default no label)
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param special_kpoints: list of tuples (str, float), place vertical lines at the given values
                            and mark them on the x-axis with the given label
    :param e_fermi: float, determines, where to put the line for the fermi energy
    :param markersize_min: minimum value used in scaling points for weight
    :param markersize_scaling: factor used in scaling points for weight
    :param outfilename: filename of the output file
    :param scale_color: bool, if True (default) the weight will be additionally shown via a colormapping
    :param line_plot: bool, if True the bandstructure will be plotted with lines
                      Here no weights are supported
    :param separate_bands: bool, if True the bandstructure will be separately plotted for each band
                           allows more specific parametrization
    :param band_index: data for which eigenvalue belongs to which band (needed for line_plot and separate_bands)
    :param copy_data: bool, if True the data argument will be copied

    Kwargs will be passed on to :py:func:`bokeh_multi_scatter()` or :py:func:`bokeh_line()`
    """
    from bokeh.transform import linear_cmap
    from bokeh.models import ColumnDataSource

    if 'size_scaling' in kwargs:
        warnings.warn('size_scaling is deprecated. Use markersize_scaling instead', DeprecationWarning)
        markersize_scaling = kwargs.pop('size_scaling')

    if 'size_min' in kwargs:
        warnings.warn('size_min is deprecated. Use markersize_min instead', DeprecationWarning)
        markersize_min = kwargs.pop('size_min')

    if isinstance(kpath, (dict, pd.DataFrame, ColumnDataSource)) or kpath is None:
        warnings.warn(
            'Passing the dataframe as first argument is deprecated. Please pass in source by the keyword data'
            'and kpath and bands as the first arguments', DeprecationWarning)
        data = kpath
        kpath = kwargs.pop('k_label', 'kpath')
        bands = kwargs.pop('eigenvalues', 'eigenvalues_up')

    if 'weight' in kwargs:
        warnings.warn('The weight argument is deprecated. Use size_data and color_data instead', DeprecationWarning)
        size_data = kwargs.pop('weight')

    plot_data = process_data_arguments(single_plot=True,
                                       data=data,
                                       kpath=kpath,
                                       bands=bands,
                                       size=size_data,
                                       color=color_data,
                                       band_index=band_index,
                                       copy_data=copy_data,
                                       use_column_source=True)

    if line_plot and size_data is not None:
        raise ValueError('Bandstructure with lines and size scaling not supported')

    if line_plot and color_data is not None:
        raise ValueError('Bandstructure with lines and color mapping not supported')

    if line_plot or separate_bands:
        if band_index is None:
            raise ValueError('The data for band indices are needed for separate_bands and line_plot')
        plot_data.group_data('band_index')
        plot_data.sort_data('kpath')

    if scale_color and size_data is not None:
        if color_data is not None:
            raise ValueError('color_data should not be provided when scale_color is True')
        plot_data.copy_data('size', 'color', rename_original=True)

    if color_data is not None:
        kwargs['color'] = plot_data.get_keys('color')

    entries = plot_data.keys(first=True)
    if entries.size is not None:
        ylimits = (-15, 15)
        if 'limits' in kwargs:
            if 'y' in kwargs['limits']:
                ylimits = kwargs['limits']['y']

        data = plot_data.values(first=True)
        mask = np.logical_and(data.bands > ylimits[0], data.bands < ylimits[1])

        weight_max = plot_data.max('size', mask=mask)

        plot_params.set_defaults(default_type='function', marker_size=entries.size)
        if scale_color:
            plot_params.set_defaults(default_type='function',
                                     color=linear_cmap(entries.color, 'Blues256', weight_max, -0.05))

        transform = lambda size: markersize_min + markersize_scaling * size / weight_max
        plot_data.apply('size', transform)
    else:
        plot_params.set_defaults(default_type='function', color='black')

    if special_kpoints is None:
        special_kpoints = []

    xticks = []
    xticklabels = {}
    for label, pos in special_kpoints:
        #if label in ('Gamma', 'g'): Latex label missing for bokeh
        #    label = r'$\Gamma$'
        if pos.is_integer():
            xticklabels[int(pos)] = label
        xticklabels[pos] = label
        xticks.append(pos)

    lines = {'horizontal': 0}
    lines['vertical'] = xticks

    limits = {'y': (-15, 15)}
    plot_params.set_defaults(default_type='function',
                             straight_lines=lines,
                             x_ticks=xticks,
                             x_ticklabels_overwrite=xticklabels,
                             figure_kwargs={
                                 'width': 1280,
                                 'height': 720
                             },
                             x_range_padding=0.0,
                             y_range_padding=0.0,
                             legend_label='Eigenvalues',
                             limits=limits)

    if line_plot:
        return bokeh_line(plot_data.get_keys('kpath'),
                          plot_data.get_keys('bands'),
                          data=plot_data.data,
                          xlabel='',
                          ylabel=ylabel,
                          title=title,
                          set_default_legend=False,
                          saveas=saveas,
                          **kwargs)
    else:
        return bokeh_multi_scatter(plot_data.get_keys('kpath'),
                                   plot_data.get_keys('bands'),
                                   data=plot_data.data,
                                   xlabel='',
                                   ylabel=ylabel,
                                   title=title,
                                   set_default_legend=False,
                                   saveas=saveas,
                                   **kwargs)


@ensure_plotter_consistency(plot_params)
def bokeh_spinpol_bands(kpath,
                        bands_up=None,
                        bands_dn=None,
                        *,
                        size_data=None,
                        color_data=None,
                        data=None,
                        xlabel='',
                        ylabel=r'E-E_F [eV]',
                        title='',
                        special_kpoints=None,
                        markersize_min=3.0,
                        markersize_scaling=10.0,
                        saveas='bands_plot',
                        scale_color=True,
                        line_plot=False,
                        separate_bands=False,
                        band_index=None,
                        copy_data=False,
                        **kwargs):
    """
    Create an interactive bandstructure plot (spinpolarized) with bokeh
    Can make a simple plot or weight the size and color of the points against a given weight

    :param kpath: arraylike or key data for the kpoint data
    :param bands_up: arraylike or key data for the eigenvalues spin-up
    :param bands_dn: arraylike or key data for the eigenvalues spin-dn
    :param size_data: arraylike or key data the weights to emphasize (optional)
    :param color_data: str or arraylike, data for the color values with a colormap (optional)
    :param data: source for the bands data (optional) of the plot (pandas Dataframe for example)
    :param xlabel: label for the x-axis (default no label)
    :param ylabel: label for the y-axis
    :param title: title of the figure
    :param special_kpoints: list of tuples (str, float), place vertical lines at the given values
                            and mark them on the x-axis with the given label
    :param e_fermi: float, determines, where to put the line for the fermi energy
    :param markersize_min: minimum value used in scaling points for weight
    :param markersize_scaling: factor used in scaling points for weight
    :param outfilename: filename of the output file
    :param scale_color: bool, if True (default) the weight will be additionally shown via a colormapping
    :param line_plot: bool, if True the bandstructure will be plotted with lines
                      Here no weights are supported
    :param separate_bands: bool, if True the bandstructure will be separately plotted for each band
                           allows more specific parametrization
    :param band_index: data for which eigenvalue belongs to which band (needed for line_plot and separate_bands)
    :param copy_data: bool, if True the data argument will be copied

    Kwargs will be passed on to :py:func:`bokeh_multi_scatter()` or :py:func:`bokeh_line()`
    """
    from bokeh.transform import linear_cmap
    from bokeh.models import ColumnDataSource

    if 'size_scaling' in kwargs:
        warnings.warn('size_scaling is deprecated. Use markersize_scaling instead', DeprecationWarning)
        markersize_scaling = kwargs.pop('size_scaling')

    if 'size_min' in kwargs:
        warnings.warn('size_min is deprecated. Use markersize_min instead', DeprecationWarning)
        markersize_min = kwargs.pop('size_min')

    if isinstance(kpath, (dict, pd.DataFrame, ColumnDataSource)) or kpath is None:
        warnings.warn(
            'Passing the dataframe as first argument is deprecated. Please pass in source by the keyword data'
            'and kpath and bands_up and bands_dn as the first arguments', DeprecationWarning)
        data = kpath
        kpath = kwargs.pop('k_label', 'kpath')
        bands_up = kwargs.pop('eigenvalues', ['eigenvalues_up', 'eigenvalues_down'])
        bands_up, bands_dn = bands_up[0], bands_up[1]

    if 'weight' in kwargs:
        warnings.warn('The weight argument is deprecated. Use size_data and color_data instead', DeprecationWarning)
        size_data = kwargs.pop('weight')

    plot_data = process_data_arguments(data=data,
                                       kpath=kpath,
                                       bands=[bands_up, bands_dn],
                                       size=size_data,
                                       color=color_data,
                                       band_index=band_index,
                                       copy_data=copy_data,
                                       use_column_source=True)

    plot_params.single_plot = False
    plot_params.num_plots = len(plot_data)

    if len(plot_data) != 2:
        raise ValueError('Wrong number of plots specified (Only 2 permitted)')

    if line_plot and size_data is not None:
        raise ValueError('Bandstructure with lines and size scaling not supported')

    if line_plot and color_data is not None:
        raise ValueError('Bandstructure with lines and color mapping not supported')

    if line_plot or separate_bands:
        if band_index is None:
            raise ValueError('The data for band indices are needed for separate_bands and line_plot')

        plot_data.group_data('band_index')
        plot_data.sort_data('kpath')

    if scale_color and size_data is not None:
        if color_data is not None:
            raise ValueError('color_data should not be provided when scale_color is True')
        plot_data.copy_data('size', 'color', rename_original=True)

    if color_data is not None:
        kwargs['color'] = plot_data.get_keys('color')

    if any(entry.size is not None for entry in plot_data.keys()):

        ylimits = (-15, 15)
        if 'limits' in kwargs:
            if 'y' in kwargs['limits']:
                ylimits = kwargs['limits']['y']

        data = plot_data.values()
        mask = [np.logical_and(col.bands > ylimits[0], col.bands < ylimits[1]) for col in data]
        weight_max = plot_data.max('size', mask=mask)

        transform = lambda size: markersize_min + markersize_scaling * size / weight_max
        plot_data.apply('size', transform)

        plot_params.set_defaults(default_type='function', marker_size=plot_data.get_keys('size'))
        if scale_color:
            plot_params.set_defaults(default_type='function',
                                     color=[
                                         linear_cmap(name, palette, weight_max, -0.05)
                                         for name, palette in zip(plot_data.get_keys('color'), ['Blues256', 'Reds256'])
                                     ])
    else:
        color = ['blue', 'red']
        plot_params.set_defaults(default_type='function', color=color)

    if special_kpoints is None:
        special_kpoints = []

    xticks = []
    xticklabels = {}
    for label, pos in special_kpoints:
        #if label in ('Gamma', 'g'): Latex label missing for bokeh
        #    label = r'$\Gamma$'
        if pos.is_integer():
            xticklabels[int(pos)] = label
        xticklabels[pos] = label
        xticks.append(pos)

    lines = {'horizontal': 0}
    lines['vertical'] = xticks

    limits = {'y': (-15, 15)}
    plot_params.set_defaults(default_type='function',
                             straight_lines=lines,
                             x_ticks=xticks,
                             x_ticklabels_overwrite=xticklabels,
                             figure_kwargs={
                                 'width': 1280,
                                 'height': 720
                             },
                             x_range_padding=0.0,
                             y_range_padding=0.0,
                             limits=limits,
                             legend_label=['Spin Up', 'Spin Down'],
                             level=[None, 'underlay'])

    if line_plot or separate_bands:
        plot_params.num_plots = len(plot_data)
        kwargs = plot_params.expand_parameters(original_length=2, **kwargs)

    if line_plot:
        return bokeh_line(plot_data.get_keys('kpath'),
                          plot_data.get_keys('bands'),
                          data=plot_data.data,
                          xlabel='',
                          ylabel=ylabel,
                          title=title,
                          set_default_legend=False,
                          saveas=saveas,
                          **kwargs)
    else:
        return bokeh_multi_scatter(plot_data.get_keys('kpath'),
                                   plot_data.get_keys('bands'),
                                   data=plot_data.data,
                                   xlabel='',
                                   ylabel=ylabel,
                                   title=title,
                                   set_default_legend=False,
                                   saveas=saveas,
                                   **kwargs)


####################################################################################################
##################################### special plots ################################################
####################################################################################################

# periodic table

# tools="pan, xpan, ypan, poly_select, tap, wheel_zoom, xwheel_zoom, ywheel_zoom, xwheel_pan, ywheel_pan,
#         box_zoom, redo, undo, reset, save, crosshair, zoom_out, xzoom_out, yzoom_out, hover"
tooltips_def_period = [('Name', '@name'), ('Atomic number', '@{atomic number}'), ('Atomic mass', '@{atomic mass}'),
                       ('CPK color', '$color[hex, swatch]:CPK'),
                       ('Electronic configuration', '@{electronic configuration}')]


def periodic_table_plot(source,
                        display_values=[],
                        display_positions=[],
                        color_value=None,
                        tooltips=tooltips_def_period,
                        title='',
                        outfilename='periodictable.html',
                        value_color_range=[None, None],
                        log_scale=0,
                        color_map=None,
                        bokeh_palette='Plasma256',
                        toolbar_location=None,
                        tools='hover',
                        blank_color='#c4c4c4',
                        blank_outsiders=[True, True],
                        include_legend=True,
                        copy_source=True,
                        legend_labels=None,
                        color_bar_title=None,
                        show=True):
    """
    Plot function for an interactive periodic table plot. Heat map and hover tool.
    source must be a panda dataframe containing, period, group,

    param source: pandas dataframe containing everything
    param tooltips: what is shown with hover tool. values have to be in source
    example:

    .. code-block:: python

        Keys of panda DF. group, period symbol and atomic number or required...
        Index([u'atomic number', u'symbol', u'name', u'atomic mass', u'CPK',
           u'electronic configuration', u'electronegativity', u'atomic radius',
           u'ion radius', u'van der Waals radius', u'IE-1', u'EA',
           u'standard state', u'bonding type', u'melting point', u'boiling point',
           u'density', u'metal', u'year discovered', u'group', u'period',
           u'rmt_mean', u'rmt_std', u'number_of_occ', u'type_color', u'c_value'],
          dtype='object')

        tooltips_def = [("Name", "@name"),
                    ("Atomic number", "@{atomic number}"),
                    ("Atomic mass", "@{atomic mass}"),
                    ("CPK color", "$color[hex, swatch]:CPK"),
                    ("Electronic configuration", "@{electronic configuration}")]

    param display_values: list of strings, have to match source. Values to be displayed on the element rectangles
    example:["rmt_mean", "rmt_std", "number_of_occ"]
    param display_positions: list of floats, length has to match display_values, At which y offset the display values should be displayed.
    """
    # TODO: solve the use of two different color bars, we just one to use a bokeh color bar and not matplotlib...
    from bokeh.io import export_png
    from bokeh.io import output_notebook, output_file
    from bokeh.sampledata.periodic_table import elements
    from bokeh.transform import dodge, factor_cmap
    from bokeh.models import Arrow, OpenHead, NormalHead, VeeHead
    from bokeh.models import Range1d, LabelSet, Label
    from bokeh.models import LinearColorMapper, LogColorMapper, ColorBar
    from bokeh.models import BasicTicker

    from bokeh.io import show as bshow
    from bokeh.plotting import figure as bokeh_fig
    from matplotlib.colors import Normalize, LogNorm, to_hex
    from matplotlib.cm import plasma  #pylint: disable=no-name-in-module
    from matplotlib.cm import ScalarMappable

    if color_map is None:
        color_map = plasma

    if len(display_values) != len(display_positions):
        raise ValueError(
            'The input lists "display_values" and "display_positions" of "periodic_table_plot" need to have same length.'
        )

    if copy_source:
        source1 = source.copy()
    else:  # inline we change the data here!
        source1 = source

    TOOLTIPS = tooltips

    # defaults
    plot_width = 1470
    plot_height = 1040
    cbar_height = 40
    cbar_width = 500
    cbar_fontsize = 12  # size of cbar labels
    cbar_standoff = 8

    # preprocessing data
    # if colors are not given in source, color
    # if source has "type_color"]

    # colors are tricky, since for the periodic table we use matplotlib, for the legend we have to use bokeh

    # Define color map called 'color_scale'
    data = source1[color_value]
    if value_color_range[0] is not None:
        mind = value_color_range[0]
    else:
        mind = min(data)  # 0.65 #
    if value_color_range[1] is not None:
        maxd = value_color_range[1]
    else:
        maxd = max(data)  # 2.81 #

    if log_scale == 0:
        color_mapper = LinearColorMapper(palette=bokeh_palette, low=mind, high=maxd)
        norm = Normalize(vmin=mind, vmax=maxd)
    elif log_scale == 1:
        for datum in data:
            if datum < 0:
                raise ValueError('Entry for element ' + datum + ' is negative but' ' log-scale is selected')
        color_mapper = LogColorMapper(palette=bokeh_palette, low=mind, high=maxd)
        norm = LogNorm(vmin=mind, vmax=maxd)
    color_scale = ScalarMappable(norm=norm, cmap=color_map).to_rgba(data, alpha=None)

    # Define color for blank entries
    default_value = None
    color_list = []
    color_values = []
    for i in range(len(source1)):
        color_list.append(blank_color)
        color_values.append(default_value)

    for i, data_element in enumerate(source1[color_value]):
        if blank_outsiders[0] and data_element < mind:
            continue
        if blank_outsiders[1] and data_element > maxd:
            continue
        color_list[i] = to_hex(color_scale[i])
        # color_values[i] = data_element

    source1['type_color'] = color_list
    # source["c_value"] = color_values

    if include_legend:
        # we copy the Be entry and display it with some text again at another spot
        be = source1[3:4].copy()
        be['group'] = '7'
        # print(be)
        source1.loc[-1] = be.values[0]
        source1.index = source1.index + 1
        source1 = source1.sort_index()
        # df.head()
    groups = [str(x) for x in range(1, 19)]
    periods = ['I', 'II', 'III', 'IV', 'V', 'VI', 'VII']
    # Plot
    p = bokeh_fig(
        title=title,
        plot_width=plot_width,
        plot_height=plot_height,  # 450,
        x_range=groups,
        y_range=list(reversed(periods)),
        tools=tools,
        toolbar_location=toolbar_location,
        tooltips=TOOLTIPS)
    r = p.rect(
        'group',
        'period',
        0.95,
        0.95,
        source=source1,
        fill_alpha=0.6,  # legend="metal",
        color='type_color')

    text_props = {'source': source1, 'text_align': 'left', 'text_baseline': 'middle'}
    x = dodge('group', -0.4, range=p.x_range)

    # The element names
    p.text(x=x,
           y=dodge('period', 0.25, range=p.y_range),
           text='symbol',
           text_font_style='bold',
           text_font_size='14pt',
           **text_props)
    p.text(x=dodge('group', 0.1, range=p.x_range),
           y=dodge('period', 0.25, range=p.y_range),
           text='atomic number',
           text_font_size='12pt',
           **text_props)

    # The values displayed on the element boxes
    for i, label in enumerate(display_values):
        p.text(x=x,
               y=dodge('period', display_positions[i], range=p.y_range),
               text=label,
               text_font_size='10pt',
               **text_props)

        # legend
        # not a real legend, but selfmade text
        # print(be.values[0])
        # print(source.loc(-1))
        # print(be['group'])
        # p.text(x=be['group'], y=dodge("period", 0.25, range=p.y_range), text="test", text_font_style="bold", text_font_size="14pt")
        if legend_labels is not None:
            label1 = legend_labels[i]
        else:
            label1 = label

        # I do not like the hardcoded positions of the legend
        legendlabel = Label(
            x=7.1,
            y=8.4 + display_positions[i],  # x_units='screen', y_units='screen',
            text=label1,
            render_mode='canvas',  # 'css',
            border_line_color='black',
            border_line_alpha=0.0,
            background_fill_color=None,
            background_fill_alpha=1.0)
        p.add_layout(legendlabel)

        legendlabelarrow = Arrow(
            x_start=7.05,
            x_end=6.7,
            y_start=8.5 + display_positions[i],
            y_end=8.5 + display_positions[i],  # x_units='screen', y_units='screen',
            line_width=2,
            end=OpenHead(line_width=2, size=4))  # 'css',
        # border_line_color='black', border_line_alpha=0.0,
        # background_fill_color=None, background_fill_alpha=1.0))
        p.add_layout(legendlabelarrow)

    # labels = LabelSet(x=70, y=80, text=display_values, level='glyph',
    #          x_offset=0, y_offset=5, render_mode='canvas')
    # p.add_layout(labels)

    p.yaxis.major_label_text_font_size = '22pt'
    p.xaxis.visible = False
    p.outline_line_color = None
    p.grid.grid_line_color = None
    p.axis.axis_line_color = None
    p.axis.major_tick_line_color = None
    p.axis.major_label_standoff = 0
    # p.legend.orientation = "horizontal"
    p.hover.renderers = [r]
    alpha = 1.0
    # add color bar
    if color_bar_title is None:
        color_bar_title = color_value
    color_bar = ColorBar(
        color_mapper=color_mapper,
        title=color_bar_title,
        title_text_font_size='12pt',
        ticker=BasicTicker(desired_num_ticks=10),
        border_line_color=None,
        background_fill_color=None,
        # 'vertical',
        label_standoff=cbar_standoff,
        location=(plot_width * 0.2, plot_height * 0.69),
        orientation='horizontal',
        scale_alpha=alpha,
        major_label_text_font_size=str(cbar_fontsize) + 'pt',
        height=cbar_height,
        width=cbar_width)

    p.add_layout(color_bar, 'center')

    # deactivate grid

    p.grid.grid_line_color = None

    # export_png(p, filename="plot.png")
    output_file(outfilename)

    if show:
        bshow(p)

    return p


######## a 2d matrix plot ##########

######### plot convergence results plot ########


@ensure_plotter_consistency(plot_params)
def plot_convergence_results(iteration, distance, total_energy, *, saveas='convergence', **kwargs):
    """
    Plot the total energy versus the scf iteration
    and plot the distance of the density versus iterations. Uses bokeh_line and bokeh_scatter

    :param iteration: list of Int
    :param distance: list of floats
    :total_energy: list of floats
    :param show: bool, if True call show

    Kwargs will be passed on to :py:func:`bokeh_line()`

    :returns grid: bokeh grid with figures
    """
    from bokeh.layouts import gridplot

    xlabel = r'Iteration'
    ylabel1 = r'Total energy difference [Htr]'
    ylabel2 = r'Distance [me/bohr^3]'
    title1 = r'Total energy difference over scf-Iterations'
    title2 = r'Convergence (log)'

    # since we make a log plot of the total_energy make sure to plot the absolute total energy
    total_energy_abs_diff = []
    for en0, en1 in zip(total_energy[:-1], total_energy[1:]):
        total_energy_abs_diff.append(abs(en1 - en0))

    source1 = pd.DataFrame({'total_energy_delta': total_energy_abs_diff, 'iterations': iteration[1:]})
    source2 = pd.DataFrame({'distance': distance, 'iterations': iteration})

    plot_params.set_defaults(default_type='function',
                             figure_kwargs={
                                 'tools': 'hover,tap,box_zoom,zoom_out,crosshair,reset,save',
                                 'tooltips': [('Iteration', '@x'), ('Total energy distance', '@y')],
                                 'active_inspect': 'hover',
                                 'plot_width': 800,
                                 'plot_height': 450,
                                 'y_axis_type': 'log',
                                 'x_axis_type': 'linear',
                             })

    if 'show' in kwargs:
        plot_params.set_parameters(show=kwargs.pop('show'))
    if 'save_plots' in kwargs:
        plot_params.set_parameters(show=kwargs.pop('save_plots'))

    with NestedPlotParameters(plot_params):
        p1 = bokeh_line(source=source1,
                        ydata='total_energy_delta',
                        xdata='iterations',
                        xlabel=xlabel,
                        ylabel=ylabel1,
                        title=title1,
                        name='delta total energy',
                        plot_points=True,
                        show=False,
                        save_plots=False,
                        **kwargs)

    plot_params.set_defaults(default_type='function',
                             figure_kwargs={'tooltips': [('Iteration', '@x'), ('Charge distance', '@y')]})

    with NestedPlotParameters(plot_params):
        p2 = bokeh_line(source=source2,
                        ydata='distance',
                        xdata='iterations',
                        xlabel=xlabel,
                        ylabel=ylabel2,
                        title=title2,
                        name='distance',
                        plot_points=True,
                        show=False,
                        save_plots=False,
                        **kwargs)

    grid = gridplot([p1, p2], ncols=2)

    plot_params.save_plot(grid, saveas)

    return grid


@ensure_plotter_consistency(plot_params)
def plot_convergence_results_m(iterations,
                               distances,
                               total_energies,
                               *,
                               link=False,
                               nodes=None,
                               modes=None,
                               plot_label=None,
                               saveas='convergence',
                               **kwargs):
    """
    Plot the total energy versus the scf iteration
    and plot the distance of the density versus iterations in a bokeh grid for several SCF results.

    :param distances: list of lists of floats
    :total_energies: list of lists of floats
    :param iterations: list of lists of Int
    :param link: bool, optional default=False:
    :param nodes: list of node uuids or pks important for links
    :param saveas1: str, optional default='t_energy_convergence', save first figure as
    :param saveas2: str, optional default='distance_convergence', save second figure as
    :param figure_kwargs: dict, optional default={'plot_width': 600, 'plot_height': 450}, gets parsed
                          to bokeh_line
    :param kwargs: further key-word arguments for bokeh_line

    :returns grid: bokeh grid with figures
    """
    from bokeh.layouts import gridplot
    from bokeh.models import ColumnDataSource

    xlabel = r'Iteration'
    ylabel1 = r'Total energy difference [Htr]'
    ylabel2 = r'Distance [me/bohr^3]'
    title1 = r'Total energy difference over scf-Iterations'
    title2 = r'Convergence (log)'

    if nodes is not None:
        if not isinstance(nodes, list):
            nodes = [nodes]

    if plot_label is not None:
        if not isinstance(plot_label, list):
            plot_label = [plot_label]

    plot_labels1 = []
    plot_labels2 = []

    data_sources = []
    data_sources2 = []

    tooltips_scatter1 = [('Calculation id', '@id'), ('Iteration', '@x'), ('Total energy difference', '@y')]
    tooltips_scatter2 = [('Calculation id', '@id'), ('Iteration', '@x'), ('Charge distance', '@y')]

    xdata = ['x'] * len(total_energies)
    ydata = ['y'] * len(total_energies)

    # since we make a log plot of the total_energy make sure to plot the absolute total energy
    for i, (iters, total_energy, distance) in enumerate(zip(iterations, total_energies, distances)):
        if nodes is not None:
            node_id = nodes[i]
        else:
            node_id = i

        total_energy_abs_diff = []
        for en0, en1 in zip(total_energy[:-1], total_energy[1:]):
            total_energy_abs_diff.append(abs(en1 - en0))

        plot_labels1.append(f'{node_id}')
        plot_labels2.append(f'{node_id}')
        data = {'y': total_energy_abs_diff, 'x': iters[1:], 'id': [node_id] * len(total_energy_abs_diff)}
        if nodes is not None:
            data['nodes_pk'] = [str(nodes[i])] * len(total_energy_abs_diff)
        if plot_label is not None:
            data['process_label'] = [plot_label[i]] * len(total_energy_abs_diff)

        datasrc = ColumnDataSource(data)
        data_sources.append(datasrc)
        data = {'y': distance, 'x': iters, 'id': [node_id] * len(distance)}
        if nodes is not None:
            data['nodes_pk'] = [str(nodes[i])] * len(distance)
        if plot_label is not None:
            data['process_label'] = [plot_label[i]] * len(distance)

        datasrc = ColumnDataSource(data)
        data_sources2.append(datasrc)

    if plot_label is not None:
        if nodes is None:
            plot_labels1 = plot_label
            plot_labels2 = plot_label
        tooltips_scatter1.append(('process label', '@process_label'))
        tooltips_scatter2.append(('process label', '@process_label'))

    if nodes is not None:
        tooltips_scatter1.append(('outpara pk', '@nodes_pk'))
        tooltips_scatter2.append(('outpara pk', '@nodes_pk'))

    plot_params.set_defaults(default_type='function',
                             figure_kwargs={
                                 'tools': 'hover,tap,box_zoom,zoom_out,crosshair,reset,save,pan',
                                 'tooltips': tooltips_scatter1,
                                 'active_inspect': 'hover',
                                 'plot_width': 800,
                                 'plot_height': 450,
                                 'y_axis_type': 'log',
                                 'x_axis_type': 'linear',
                             },
                             legend_outside_plot_area=True)
    if 'show' in kwargs:
        plot_params.set_parameters(show=kwargs.pop('show'))
    if 'save_plots' in kwargs:
        plot_params.set_parameters(show=kwargs.pop('save_plots'))

    # plot
    with NestedPlotParameters(plot_params):
        p1 = bokeh_line(source=data_sources,
                        ydata=ydata,
                        xdata=xdata,
                        xlabel=xlabel,
                        ylabel=ylabel1,
                        title=title1,
                        name=plot_labels1,
                        legend_label=plot_labels1,
                        plot_points=True,
                        show=False,
                        save_plots=False,
                        **kwargs)

    plot_params.set_defaults(default_type='function', figure_kwargs={
        'tooltips': tooltips_scatter2,
    })

    with NestedPlotParameters(plot_params):
        p2 = bokeh_line(source=data_sources2,
                        ydata=ydata,
                        xdata=xdata,
                        xlabel=xlabel,
                        ylabel=ylabel2,
                        title=title2,
                        name=plot_labels2,
                        legend_label=plot_labels2,
                        plot_points=True,
                        show=False,
                        save_plots=False,
                        **kwargs)

    grid = gridplot([p1, p2], ncols=2)

    plot_params.save_plot(grid, saveas)

    return grid


######### plot_convex_hull plot ########


def plot_convex_hull2d(hull,
                       title='Convex Hull',
                       xlabel='Atomic Procentage',
                       ylabel='Formation energy / atom [eV]',
                       linestyle='-',
                       marker='o',
                       legend=True,
                       legend_option={},
                       saveas='convex_hull',
                       limits=[None, None],
                       scale=[None, None],
                       axis=None,
                       color='k',
                       color_line='k',
                       linewidth=2,
                       markersize=8,
                       marker_hull='o',
                       markersize_hull=8,
                       **kwargs):
    """
    Plot method for a 2d convex hull diagram

    :param hull: scipy.spatial.ConvexHull
    """
    pass



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Contains utility to add new pairs of input/output schemas.
"""
from masci_tools.util.xml.common_functions import clear_xml
import os
import sys
import shutil
from lxml import etree

PACKAGE_DIRECTORY = os.path.dirname(os.path.abspath(__file__))


def add_fleur_schema(path, overwrite=False):
    """
    Adds the FleurInput/OutputSchema from the specified path (folder containing the Schemas)
    to the folder with the correct version number and creates the schema_dicts

    :param path: path to the folder containing the schema files
    :param overwrite: bool, if True and the schema with the same version exists it will be
                      overwritten. Otherwise an error is raised
    """

    schema_path = os.path.join(path, 'FleurInputSchema.xsd')
    if os.path.isfile(schema_path):
        xmlschema = etree.parse(schema_path)
        xmlschema, _ = clear_xml(xmlschema)

        namespaces = {'xsd': 'http://www.w3.org/2001/XMLSchema'}
        inp_version = xmlschema.xpath('/xsd:schema/@version', namespaces=namespaces)[0]

        copy_schema_folder = os.path.abspath(os.path.join(PACKAGE_DIRECTORY, f'./{inp_version}'))
        copy_schema_file = os.path.abspath(os.path.join(copy_schema_folder, 'FleurInputSchema.xsd'))
        if os.path.isfile(copy_schema_file) and not overwrite:
            raise ValueError(
                f'Input Schema for version {inp_version} already exists. Use overwrite=True to replace the Schema')

        os.makedirs(copy_schema_folder, exist_ok=True)
        if not os.path.isfile(os.path.abspath(os.path.join(copy_schema_folder, '__init__.py'))):
            with open(os.path.abspath(os.path.join(copy_schema_folder, '__init__.py')), 'w') as f:
                pass
        shutil.copy(schema_path, copy_schema_file)

    schema_path = os.path.join(path, 'FleurOutputSchema.xsd')
    if os.path.isfile(schema_path):
        xmlschema = etree.parse(schema_path)
        xmlschema, _ = clear_xml(xmlschema)

        namespaces = {'xsd': 'http://www.w3.org/2001/XMLSchema'}
        out_version = xmlschema.xpath('/xsd:schema/@version', namespaces=namespaces)[0]

        copy_schema_folder = os.path.abspath(os.path.join(PACKAGE_DIRECTORY, f'./{out_version}'))
        copy_schema_file = os.path.abspath(os.path.join(copy_schema_folder, 'FleurOutputSchema.xsd'))
        if os.path.isfile(copy_schema_file) and not overwrite:
            raise ValueError(
                f'Output Schema for version {out_version} already exists. Use overwrite=True to replace the Schema')

        os.makedirs(copy_schema_folder, exist_ok=True)
        if not os.path.isfile(os.path.abspath(os.path.join(copy_schema_folder, '__init__.py'))):
            with open(os.path.abspath(os.path.join(copy_schema_folder, '__init__.py')), 'w') as f:
                pass
        shutil.copy(schema_path, copy_schema_file)


if __name__ == '__main__':
    add_fleur_schema(sys.argv[1], overwrite=sys.argv[2])



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module defines commonly used recipes for the :py:class:`~masci_tools.io.parsers.hdf5.reader.HDF5Reader`

Available are:
    - Recipe for bandstructure calculations with Fleur
    - Recipes for almost all DOS calculation modes of Fleur

A Recipe is a python dictionary in a specific format.

A Template Example:

.. code-block:: python

    from masci_tools.io.parser.hdf5.readers import Transformation, AttribTransformation

    RecipeExample = {
        'datasets': {
            'example_dataset': {
                'h5path': '/path/in/hdf/file',
                'transforms': [Transformation(name='get_first_element')]
            },
            'example_attrib_transform': {
                'h5path': '/other/path/in/hdf/file',
                'transforms': [AttribTransformation(name='multiply_by_attribute', attrib_name='example_attribute')]
            }
        },
        'attributes': {
            'example_attribute': {
                'h5path':
                '/path/in/hdf/file',
                'transforms':
                [Transformation(name='get_attribute', args=('attribName',)),
                 Transformation(name='get_first_element')]
            }
        }
    }

The Recipe consists of two sections 'datasets' and 'attributes'. All data from these two sections will be returned
in separate python dictionaries by the :py:class:`~masci_tools.io.parsers.hdf5.reader.HDF5Reader` class

Each entry in those sections has to have a `h5path` entry, which will specify the dataset to initially
read from the hdf file. Then each entry can define a entry `transforms` with a list of the namedtuples
imported at the top of the code example. These correponds to function calls to functions in
:py:mod:`~masci_tools.io.parsers.hdf5.transforms` to transform the read in data

Entries in the `attributes` section are read and transformed first and can subsequently be used in transformations
for the `datasets`. These correpsond to the transforms created with the :py:class:`~masci_tools.io.parsers.hdf5.reader.AttribTransformation`
namedtuple instead of :py:class:`~masci_tools.io.parsers.hdf5.reader.Transformation`.
"""
from masci_tools.util.constants import HTR_TO_EV, BOHR_A
from .reader import Transformation, AttribTransformation


def dos_recipe_format(group):
    """
    Format for denisty of states calculations retrieving the DOS from the given group

    :param group: str of the group the DOS should be taken from

    :returns: dict of the recipe to retrieve a DOS calculation
    """

    if group == 'Local':
        atom_prefix = 'MT:'
    elif group == 'jDOS':
        atom_prefix = 'jDOS:'
    elif group == 'Orbcomp':
        atom_prefix = 'ORB:'
    elif group == 'MCD':
        atom_prefix = 'At'
    else:
        raise ValueError(f'Unknown group: {group}')

    return {
        'datasets': {
            'dos': {
                'h5path':
                f'/{group}/DOS',
                'transforms': [
                    Transformation(name='get_all_child_datasets', args=(), kwargs={'ignore': 'energyGrid'}),
                    AttribTransformation(name='add_partial_sums',
                                         attrib_name='atoms_groups',
                                         args=(f'{atom_prefix}{{}}'.format,),
                                         kwargs={'make_set': True}),
                    Transformation(name='multiply_scalar', args=(1.0 / HTR_TO_EV,), kwargs={}),
                    Transformation(
                        name='split_array',
                        args=(),
                        kwargs={'suffixes': ['up', 'down']},
                    )
                ],
                'unpack_dict':
                True,
            },
            'energy_grid': {
                'h5path': f'/{group}/DOS/energyGrid',
                'transforms': [Transformation(name='multiply_scalar', args=(HTR_TO_EV,), kwargs={})]
            }
        },
        'attributes': {
            'group_name': {
                'h5path': f'/{group}',
                'transforms': [
                    Transformation(name='get_name', args=(), kwargs={}),
                ],
            },
            'n_types': {
                'h5path':
                '/atoms',
                'description':
                'Number of atom types',
                'transforms': [
                    Transformation(name='get_attribute', args=('nTypes',), kwargs={}),
                    Transformation(name='get_first_element', args=(), kwargs={})
                ]
            },
            'atoms_elements': {
                'h5path': '/atoms/atomicNumbers',
                'description': 'Atomic numbers',
                'transforms': [Transformation(name='periodic_elements', args=(), kwargs={})]
            },
            'atoms_groups': {
                'h5path': '/atoms/equivAtomsGroup'
            },
            'fermi_energy': {
                'h5path':
                '/general',
                'description':
                'fermi_energy of the system',
                'transforms': [
                    Transformation(name='get_attribute', args=('lastFermiEnergy',), kwargs={}),
                    Transformation(name='get_first_element', args=(), kwargs={})
                ]
            },
            'spins': {
                'h5path':
                '/general',
                'description':
                'number of distinct spin directions in the system',
                'transforms': [
                    Transformation(name='get_attribute', args=('spins',), kwargs={}),
                    Transformation(name='get_first_element', args=(), kwargs={})
                ]
            }
        }
    }


#DOS Recipes
FleurDOS = dos_recipe_format('Local')
FleurJDOS = dos_recipe_format('jDOS')
FleurORBCOMP = dos_recipe_format('Orbcomp')
FleurMCD = dos_recipe_format('MCD')


def bands_recipe_format(group, simple=False):
    """
    Format for bandstructure calculations retrieving weights from the given group

    :param group: str of the group the weights should be taken from
    :param simple: bool, if True no additional weights are retrieved with the produced recipe

    :returns: dict of the recipe to retrieve a bandstructure calculation
    """

    if group == 'Local':
        atom_prefix = 'MT:'
    elif group == 'jDOS':
        atom_prefix = 'jDOS:'
    elif group == 'Orbcomp':
        atom_prefix = 'ORB:'
    elif group == 'MCD':
        atom_prefix = 'At'
    else:
        raise ValueError(f'Unknown group: {group}')

    recipe = {
        'datasets': {
            'eigenvalues': {
                'h5path':
                f'/{group}/BS/eigenvalues',
                'transforms': [
                    AttribTransformation(name='shift_by_attribute',
                                         attrib_name='fermi_energy',
                                         args=(),
                                         kwargs={
                                             'negative': True,
                                         }),
                    Transformation(name='multiply_scalar', args=(HTR_TO_EV,), kwargs={}),
                    Transformation(name='split_array',
                                   args=(),
                                   kwargs={
                                       'suffixes': ['up', 'down'],
                                       'name': 'eigenvalues'
                                   }),
                    Transformation(name='flatten_array', args=(), kwargs={})
                ],
                'unpack_dict':
                True
            },
            'kpath': {
                'h5path':
                '/kpts/coordinates',
                'transforms': [
                    AttribTransformation(name='multiply_by_attribute',
                                         attrib_name='reciprocal_cell',
                                         args=(),
                                         kwargs={'transpose': True}),
                    Transformation(name='calculate_norm', args=(), kwargs={'between_neighbours': True}),
                    Transformation(name='cumulative_sum', args=(), kwargs={}),
                    AttribTransformation(name='repeat_array_by_attribute', attrib_name='nbands', args=(), kwargs={}),
                ]
            },
        },
        'attributes': {
            'group_name': {
                'h5path': f'/{group}',
                'transforms': [
                    Transformation(name='get_name', args=(), kwargs={}),
                ],
            },
            'kpoints': {
                'h5path': '/kpts/coordinates',
            },
            'nkpts': {
                'h5path':
                '/Local/BS/eigenvalues',
                'transforms': [
                    Transformation(name='get_shape', args=(), kwargs={}),
                    Transformation(name='index_dataset', args=(1,), kwargs={})
                ]
            },
            'nbands': {
                'h5path':
                '/Local/BS/eigenvalues',
                'transforms': [
                    Transformation(name='get_shape', args=(), kwargs={}),
                    Transformation(name='index_dataset', args=(2,), kwargs={})
                ]
            },
            'atoms_elements': {
                'h5path': '/atoms/atomicNumbers',
                'description': 'Atomic numbers',
                'transforms': [Transformation(name='periodic_elements', args=(), kwargs={})]
            },
            'n_types': {
                'h5path':
                '/atoms',
                'description':
                'Number of atom types',
                'transforms': [
                    Transformation(name='get_attribute', args=('nTypes',), kwargs={}),
                    Transformation(name='get_first_element', args=(), kwargs={})
                ]
            },
            'atoms_position': {
                'h5path': '/atoms/positions',
                'description': 'Atom coordinates per atom',
            },
            'atoms_groups': {
                'h5path': '/atoms/equivAtomsGroup'
            },
            'reciprocal_cell': {
                'h5path': '/cell/reciprocalCell'
            },
            'bravais_matrix': {
                'h5path': '/cell/bravaisMatrix',
                'description': 'Coordinate transformation internal to physical for atoms',
                'transforms': [Transformation(name='multiply_scalar', args=(BOHR_A,), kwargs={})]
            },
            'special_kpoint_indices': {
                'h5path': '/kpts/specialPointIndices',
                'transforms': [Transformation(name='shift_dataset', args=(-1,), kwargs={})]
            },
            'special_kpoint_labels': {
                'h5path': '/kpts/specialPointLabels',
                'transforms': [Transformation(name='convert_to_str', args=(), kwargs={})]
            },
            'fermi_energy': {
                'h5path':
                '/general',
                'description':
                'fermi_energy of the system',
                'transforms': [
                    Transformation(name='get_attribute', args=('lastFermiEnergy',), kwargs={}),
                    Transformation(name='get_first_element', args=(), kwargs={})
                ]
            },
            'spins': {
                'h5path':
                '/general',
                'description':
                'number of distinct spin directions in the system',
                'transforms': [
                    Transformation(name='get_attribute', args=('spins',), kwargs={}),
                    Transformation(name='get_first_element', args=(), kwargs={})
                ]
            }
        }
    }

    if simple:
        return recipe

    recipe['datasets']['weights'] = {
        'h5path':
        f'/{group}/BS',
        'transforms': [
            Transformation(name='get_all_child_datasets', args=(), kwargs={'ignore': ['eigenvalues', 'kpts']}),
            AttribTransformation(name='add_partial_sums',
                                 attrib_name='atoms_groups',
                                 args=(f'{atom_prefix}{{}}'.format,),
                                 kwargs={'make_set': True}),
            Transformation(name='split_array', args=(), kwargs={'suffixes': ['up', 'down']}),
            Transformation(name='flatten_array', args=(), kwargs={})
        ],
        'unpack_dict':
        True
    }

    return recipe


def get_fleur_bands_specific_weights(weight_name, group='Local'):
    """
    Recipe for bandstructure calculations only retrieving one
    additional weight besides the eigenvalues and kpath

    :param weight_name: key or list of keys of the weight(s) to retrieve
    :param group: optional str (default Local) name of the group from where to take the weights

    :returns: dict of the recipe to retrieve a simple bandstructure
              plus the one specified weight
    """
    recipe = bands_recipe_format(group, simple=True)

    if isinstance(weight_name, str):
        weight_name = [weight_name]

    for name in weight_name:
        if 'MT:' in name and not any(c in name for c in ('s', 'p', 'd', 'f')):
            #We need to get all orbitals and calculate the atom sum
            recipe['datasets'][name] = {
                'h5path':
                f'/{group}/BS',
                'transforms': [
                    Transformation(name='get_all_child_datasets', args=(), kwargs={'contains': name}),
                    Transformation(name='sum_over_dict_entries', args=(), kwargs={'overwrite_dict': True}),
                    Transformation(name='split_array', args=(), kwargs={
                        'suffixes': ['up', 'down'],
                        'name': name
                    }),
                    Transformation(name='flatten_array', args=(), kwargs={})
                ],
                'unpack_dict':
                True
            }
        else:
            recipe['datasets'][name] = {
                'h5path':
                f'/{group}/BS/{name}',
                'transforms': [
                    Transformation(name='split_array', args=(), kwargs={
                        'suffixes': ['up', 'down'],
                        'name': name
                    }),
                    Transformation(name='flatten_array', args=(), kwargs={})
                ],
                'unpack_dict':
                True
            }

    return recipe


#Recipe for bandstructures
FleurBands = bands_recipe_format('Local')
FleurOrbcompBands = bands_recipe_format('Orbcomp')
FleurjDOSBands = bands_recipe_format('jDOS')
FleurMCDBands = bands_recipe_format('MCD')
#Recipe for bandstructures without reading in weights
FleurSimpleBands = bands_recipe_format('Local', simple=True)



# -*- coding: utf-8 -*-
"""
Fixtures for testing bokeh plots.

The actual rendering is not tested. Instead the produced json is put through
the pytest-regressions plugin to be notified when it changes
"""
import pytest
import re
from pathlib import Path
import os
import warnings

pytest_plugins = ('pytest_regressions',)

__all__ = ('pytest_addoption', 'check_bokeh_plot')


def pytest_addoption(parser):
    parser.addoption('--show-bokeh-plots', action='store_true', help='Show the produced bokeh plots')
    parser.addoption('--add-bokeh-version',
                     action='store_true',
                     help='Store the results for a new installed bokeh version')


@pytest.fixture(scope='function')
def check_bokeh_plot(data_regression, clean_bokeh_json, pytestconfig, bokeh_basename, previous_bokeh_results, datadir):

    def _regression_bokeh_plot(bokeh_fig):

        basename = bokeh_basename()  #This will skip the test if bokeh is not installed

        if pytestconfig.getoption('--show-bokeh-plots'):
            from bokeh.io import show
            show(bokeh_fig)
        else:
            if not (datadir / basename).parent.is_dir():
                filename = basename.name
                prev_version = previous_bokeh_results(basename.parent)

                if not pytestconfig.getoption('--add-bokeh-version'):
                    if prev_version is not None:
                        basename = Path(prev_version) / filename
                        warnings.warn(f'Results for bokeh version {basename.parent} not available.'
                                      f'Using the last available version {prev_version}'
                                      'Use the option --add-bokeh-version to add results for this version')

            from bokeh.io import curdoc

            curdoc().clear()
            curdoc().add_root(bokeh_fig)
            data_regression.check(clean_bokeh_json(curdoc().to_json()), basename=os.fspath(basename))

    return _regression_bokeh_plot


@pytest.fixture(scope='function', name='bokeh_basename')
def fixture_bokeh_basename(request):

    def _get_bokeh_basename():
        try:
            import bokeh
            current_bokeh_version = bokeh.__version__
        except ImportError:
            pytest.skip('Bokeh regression tests are skipped only exectued if bokeh is installed')

        #Copied from pytest-regressions
        basename = re.sub(r'[\W]', '_', request.node.name)
        return Path(f'bokeh-{current_bokeh_version}') / basename

    return _get_bokeh_basename


@pytest.fixture(scope='function', name='previous_bokeh_results')
def fixture_previous_bokeh_results(datadir):

    def _get_previous_test_results(current_version):

        latest_version = (0, 0, 0)

        for f in os.scandir(datadir):
            if f.is_dir() and f.name.startswith('bokeh-'):
                _, _, version = f.name.partition('-')
                version = tuple(int(x) for x in version.split('.'))
                if latest_version < version <= current_version:
                    latest_version = version
        if latest_version > (0, 0, 0):
            return '.'.join(map(str, latest_version))
        else:
            return None

    return _get_previous_test_results


@pytest.fixture(scope='function', name='clean_bokeh_json')
def fixture_clean_bokeh_json():
    """
    Make the dict form the produced json data
    suitable for data_regression

    - remove any reference to ids
    - sort lists after types and given attributes for reproducible order

    :param data: dict with the json data produced for the bokeh figure
    """

    def _clean_bokeh_json(data, np_precision=5, data_entry=False):
        """
        Make the dict form the produced json data
        suitable for data_regression

        - remove any reference to ids
        - sort lists after types and given attributes for reproducible order

        :param data: dict with the json data produced for the bokeh figure
        """
        from bokeh.util.serialization import decode_base64_dict, encode_base64_dict
        import numpy as np

        def get_contained_keys(dict_val):

            keys = set(dict_val.keys())
            for key, val in dict_val.items():
                if isinstance(val, dict):
                    keys = keys.union(get_contained_keys(val))

            return keys

        def get_normalized_order(dict_val, key_order, normed=None):

            if normed is None:
                normed = [(key, ()) for key in key_order]
            for key, val in dict_val.items():
                if isinstance(val, dict):
                    normed = get_normalized_order(val, key_order, normed=normed)
                elif key != 'type':
                    index = key_order.index(key)
                    if isinstance(val, list):
                        normed[index] += (tuple(val),)
                    elif isinstance(val, float):
                        normed[index] += (str(val),)
                    else:
                        normed[index] += (val,)
            return normed

        def normalize_list_of_dicts(list_of_dicts):

            contained_keys = set()
            for data in list_of_dicts:
                contained_keys = contained_keys.union(get_contained_keys(data))
            contained_keys.discard('type')

            return sorted(list_of_dicts,
                          key=lambda x: (x['type'], *tuple(get_normalized_order(x, sorted(contained_keys)))))

        if '__ndarray__' in data:
            array = decode_base64_dict(data)
            array = np.around(array, decimals=np_precision)
            data = encode_base64_dict(array)

        for key, val in list(data.items()):
            if key in ('id', 'root_ids'):
                data.pop(key)
            elif isinstance(val, dict):
                data[key] = _clean_bokeh_json(val, np_precision=np_precision, data_entry=key == 'data')
            elif isinstance(val, list):
                for index, entry in enumerate(val):
                    if isinstance(entry, dict):
                        val[index] = _clean_bokeh_json(entry)
                if all(isinstance(x, dict) for x in val):
                    data[key] = normalize_list_of_dicts(val)
                elif all(isinstance(x, int) for x in val) and data_entry:
                    data[key] = encode_base64_dict(np.array(val))
                else:
                    data[key] = val

        return data

    return _clean_bokeh_json



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module provides functions to extract distinct parts of the fleur xml files
for easy versioning and reuse
"""
from masci_tools.io.parsers.fleur.fleur_schema import schema_dict_version_dispatch
from lxml import etree
import warnings


def get_fleur_modes(xmltree, schema_dict, logger=None):
    """
    Determine the calculation modes of fleur for the given xml file. Calculation modes
    are things that change the produced files or output in the out.xml files

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors

    :returns: dictionary with all the extracted calculation modes

    The following modes are inspected:

        - `jspin`: How many spins are considered in the calculation
        - `noco`: Is the calculation non-collinear?
        - `soc`: Is spin-orbit coupling included?
        - `relax`: Is the calculation a structure relaxation?
        - `gw`: Special mode for GW/Spex calculations
        - `force_theorem`: Is a Force theorem calculation performed?
        - `film`: Is the structure a film system
        - `ldau`: Is LDA+U included?
        - `dos`: Is it a density of states calculation?
        - `band`: Is it a bandstructure calculation?
        - `bz_integration`: How is the integration over the Brillouin-Zone performed?

    """
    from masci_tools.util.schema_dict_util import read_constants
    from masci_tools.util.schema_dict_util import evaluate_attribute, tag_exists
    from masci_tools.util.xml.common_functions import clear_xml

    if isinstance(xmltree, etree._ElementTree):
        xmltree, _ = clear_xml(xmltree)
        root = xmltree.getroot()
    else:
        root = xmltree
    constants = read_constants(root, schema_dict)

    fleur_modes = {}
    fleur_modes['jspin'] = evaluate_attribute(root, schema_dict, 'jspins', logger=logger, constants=constants)

    noco = evaluate_attribute(root, schema_dict, 'l_noco', constants=constants, logger=logger, optional=True)
    if noco is None:
        noco = False
    fleur_modes['noco'] = noco

    soc = evaluate_attribute(root, schema_dict, 'l_soc', constants=constants, logger=logger, optional=True)
    if soc is None:
        soc = False
    fleur_modes['soc'] = soc

    relax = evaluate_attribute(root, schema_dict, 'l_f', constants=constants, logger=logger, optional=True)
    if relax is None:
        relax = False
    fleur_modes['relax'] = relax

    gw = None
    if tag_exists(root, schema_dict, 'expertModes', logger=logger):
        gw = evaluate_attribute(root, schema_dict, 'gw', constants=constants, logger=logger, optional=True)
        if gw is None and schema_dict.inp_version >= (0, 34):
            gw = evaluate_attribute(root, schema_dict, 'spex', constants=constants, logger=logger, optional=True)

    if gw is None:
        gw = False
    else:
        gw = gw != 0
    fleur_modes['gw'] = gw

    if schema_dict.inp_version > (0, 27):
        fleur_modes['force_theorem'] = tag_exists(root, schema_dict, 'forceTheorem', logger=logger)
    else:
        fleur_modes['force_theorem'] = False

    if schema_dict.inp_version >= (0, 33):
        if tag_exists(root, schema_dict, 'cFCoeffs', logger=logger):
            cf_coeff = any(
                evaluate_attribute(root,
                                   schema_dict,
                                   'potential',
                                   contains='cFCoeffs',
                                   logger=logger,
                                   list_return=True,
                                   optional=True))
            cf_coeff = cf_coeff or any(
                evaluate_attribute(root,
                                   schema_dict,
                                   'chargeDensity',
                                   contains='cFCoeffs',
                                   logger=logger,
                                   list_return=True,
                                   optional=True))
        else:
            cf_coeff = False
        fleur_modes['cf_coeff'] = cf_coeff
    else:
        fleur_modes['cf_coeff'] = False

    plot = None
    if tag_exists(root, schema_dict, 'plotting', logger=logger):
        plot = evaluate_attribute(root, schema_dict, 'iplot', logger=logger, optional=True)

    if schema_dict.inp_version >= (0, 29) and plot is not None:
        plot = isinstance(plot, int) and plot != 0

    if plot is None:
        plot = False
    fleur_modes['plot'] = plot

    fleur_modes['film'] = tag_exists(root, schema_dict, 'filmPos', logger=logger)
    fleur_modes['ldau'] = tag_exists(root, schema_dict, 'ldaU', contains='species', logger=logger)
    fleur_modes['dos'] = evaluate_attribute(root, schema_dict, 'dos', constants=constants, logger=logger)
    fleur_modes['band'] = evaluate_attribute(root, schema_dict, 'band', constants=constants, logger=logger)
    fleur_modes['bz_integration'] = evaluate_attribute(root,
                                                       schema_dict,
                                                       'mode',
                                                       constants=constants,
                                                       tag_name='bzIntegration',
                                                       logger=logger)

    greensf = False
    if schema_dict.inp_version >= (0, 32):
        #We make the assumption that the existence of a greensfCalculation
        #tag implies the existence of a greens function calculation
        greensf = tag_exists(root, schema_dict, 'greensfCalculation', contains='species', logger=logger)
        greensf = greensf or tag_exists(root, schema_dict, 'torgueCalculation', contains='species', logger=logger)
    fleur_modes['greensf'] = greensf

    ldahia = False
    if schema_dict.inp_version >= (0, 32):
        ldahia = tag_exists(root, schema_dict, 'ldaHIA', contains='species', logger=logger)
    fleur_modes['ldahia'] = ldahia

    return fleur_modes


@schema_dict_version_dispatch(output_schema=False)
def get_nkpts(xmltree, schema_dict, logger=None):
    """
    Get the number of kpoints that will be used in the calculation specified in the given
    fleur XMl file.

    .. warning::
        For file versions before Max5 only kPointList or kPointCount tags will work. However,
        for kPointCount there is no real guarantee that for every occasion it will correspond
        to the number of kpoints. So a warning is written out

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors

    :returns: int with the number of kpoints
    """
    from masci_tools.util.schema_dict_util import eval_simple_xpath
    from masci_tools.util.schema_dict_util import evaluate_attribute
    from masci_tools.util.xml.common_functions import clear_xml

    if isinstance(xmltree, etree._ElementTree):
        xmltree, _ = clear_xml(xmltree)
        root = xmltree.getroot()
    else:
        root = xmltree

    #Get the name of the current selected kPointSet
    list_name = evaluate_attribute(root, schema_dict, 'listName', logger=logger)

    kpointlists = eval_simple_xpath(root, schema_dict, 'kPointList', list_return=True, logger=logger)

    if len(kpointlists) == 0:
        raise ValueError('No Kpoint lists found in the given inp.xml')

    labels = [kpoint_set.attrib.get('name') for kpoint_set in kpointlists]
    if list_name not in labels:
        raise ValueError(f'Selected Kpoint list with the name: {list_name} does not exist'
                         f'Available list names: {labels}')

    kpoint_index = labels.index(list_name)

    kpoint_set = kpointlists[kpoint_index]

    nkpts = evaluate_attribute(kpoint_set, schema_dict, 'count', logger=logger)

    return nkpts


@get_nkpts.register(max_version='0.31')
def get_nkpts_max4(xmltree, schema_dict, logger=None):
    """
    Get the number of kpoints that will be used in the calculation specified in the given
    fleur XMl file. Version specific for Max4 versions or older

    .. warning::
        For file versions before Max5 only kPointList or kPointCount tags will work. However,
        for kPointCount there is no real guarantee that for every occasion it will correspond
        to the number of kpoints. So a warning is written out

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors

    :returns: int with the number of kpoints
    """
    from masci_tools.util.schema_dict_util import evaluate_attribute, eval_simple_xpath
    from masci_tools.util.xml.common_functions import clear_xml

    if isinstance(xmltree, etree._ElementTree):
        xmltree, _ = clear_xml(xmltree)
        root = xmltree.getroot()
    else:
        root = xmltree

    modes = get_fleur_modes(root, schema_dict, logger=logger)

    alt_kpt_set = None
    if modes['band'] or modes['gw']:
        expected_mode = 'bands' if modes['band'] else 'gw'
        alt_kpts = eval_simple_xpath(root, schema_dict, 'altKPointSet', list_return=True, logger=logger)
        for kpt_set in alt_kpts:
            if evaluate_attribute(kpt_set, schema_dict, 'purpose', logger=logger) == expected_mode:
                alt_kpt_set = kpt_set
                break

    kpt_tag = None
    if alt_kpt_set is not None:
        kpt_tag = eval_simple_xpath(alt_kpt_set, schema_dict, 'kPointList', list_return=True, logger=logger)
        if len(kpt_tag) == 0:
            kpt_tag = eval_simple_xpath(alt_kpt_set, schema_dict, 'kPointCount', list_return=True, logger=logger)
            if len(kpt_tag) == 0:
                kpt_tag = None
            else:
                warnings.warn('kPointCount is not guaranteed to result in the given number of kpoints')

    if kpt_tag is None:
        kpt_tag = eval_simple_xpath(root,
                                    schema_dict,
                                    'kPointList',
                                    not_contains='altKPointSet',
                                    list_return=True,
                                    logger=logger)
        if len(kpt_tag) == 0:
            kpt_tag = eval_simple_xpath(root,
                                        schema_dict,
                                        'kPointCount',
                                        not_contains='altKPointSet',
                                        list_return=True,
                                        logger=logger)
            if len(kpt_tag) == 0:
                raise ValueError('No kPointList or kPointCount found')
            else:
                warnings.warn('kPointCount is not guaranteed to result in the given number of kpoints')

    kpt_tag = kpt_tag[0]

    nkpts = evaluate_attribute(kpt_tag, schema_dict, 'count', logger=logger)

    return nkpts


def get_cell(xmltree, schema_dict, logger=None, convert_to_angstroem=True):
    """
    Get the Bravais matrix from the given fleur xml file. In addition a list
    determining in, which directions there are periodic boundary conditions
    in the system.

    .. warning::
        Only the explicit definition of the Bravais matrix is supported.
        Old inputs containing the `latnam` definitions are not supported

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors
    :param convert_to_angstroem: bool if True the bravais matrix is converted to angstroem

    :returns: numpy array of the bravais matrix and list of boolean values for
              periodic boundary conditions
    """
    from masci_tools.util.schema_dict_util import read_constants, eval_simple_xpath
    from masci_tools.util.schema_dict_util import evaluate_text, tag_exists, evaluate_attribute
    from masci_tools.util.xml.common_functions import clear_xml
    from masci_tools.util.constants import BOHR_A
    import numpy as np

    if isinstance(xmltree, etree._ElementTree):
        xmltree, _ = clear_xml(xmltree)
        root = xmltree.getroot()
    else:
        root = xmltree
    constants = read_constants(root, schema_dict, logger=logger)

    cell = None
    lattice_tag = None
    if tag_exists(root, schema_dict, 'bulkLattice', logger=logger):
        lattice_tag = eval_simple_xpath(root, schema_dict, 'bulkLattice', logger=logger)
        pbc = [True, True, True]
    elif tag_exists(root, schema_dict, 'filmLattice', logger=logger):
        lattice_tag = eval_simple_xpath(root, schema_dict, 'filmLattice', logger=logger)
        pbc = [True, True, False]

    if lattice_tag is not None:
        lattice_scale = evaluate_attribute(lattice_tag,
                                           schema_dict,
                                           'scale',
                                           constants=constants,
                                           logger=logger,
                                           not_contains={'/a', 'c/'})

        row1 = evaluate_text(lattice_tag,
                             schema_dict,
                             'row-1',
                             constants=constants,
                             contains='bravaisMatrix',
                             logger=logger,
                             optional=True)
        row2 = evaluate_text(lattice_tag,
                             schema_dict,
                             'row-2',
                             constants=constants,
                             contains='bravaisMatrix',
                             logger=logger,
                             optional=True)
        row3 = evaluate_text(lattice_tag,
                             schema_dict,
                             'row-3',
                             constants=constants,
                             contains='bravaisMatrix',
                             logger=logger,
                             optional=True)

        if all(x is not None and x != [] for x in [row1, row2, row3]):
            cell = np.array([row1, row2, row3]) * lattice_scale
            if convert_to_angstroem:
                cell *= BOHR_A

    if cell is None:
        raise ValueError('Could not extract Bravais matrix out of inp.xml. Is the '
                         'Bravais matrix explicitly given? i.e Latnam definition '
                         'not supported.')

    return cell, pbc


def _get_species_info(xmltree, schema_dict, logger=None):
    """
    Gets the species identifiers and information.
    Used to keep species information consistent between
    :py:func:`get_parameter_data` and :py:func:`get_structure_data`

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors

    :returns: Tuple of dicts, containing the normalized species ids
              and the elements for each species
    """
    from masci_tools.util.xml.common_functions import clear_xml
    from masci_tools.util.schema_dict_util import read_constants, evaluate_attribute
    import re

    if isinstance(xmltree, etree._ElementTree):
        xmltree, _ = clear_xml(xmltree)
        root = xmltree.getroot()
    else:
        root = xmltree
    constants = read_constants(root, schema_dict, logger=logger)

    names = evaluate_attribute(root,
                               schema_dict,
                               'name',
                               constants=constants,
                               contains='species',
                               logger=logger,
                               list_return=True)
    elements = evaluate_attribute(root,
                                  schema_dict,
                                  'element',
                                  constants=constants,
                                  contains='species',
                                  logger=logger,
                                  list_return=True)

    if len(names) != len(elements):
        raise ValueError(
            f'Failed to read in species names and elements. Got {len(names)} names and {len(elements)} elements')

    species_info = {}
    for name, element in zip(names, elements):
        #Check if the species name has a numerical id at the end (separated by - or .)
        #And add all of them first
        species_info[name] = {}
        species_info[name]['element'] = element
        species_info[name]['normed_name'] = name
        match = re.fullmatch(r'(.+[\-\.])([1-9]+)', name)
        if match:
            species_info[name]['id'] = int(match.group(2))

    for name, info in species_info.items():
        if 'id' not in info:
            element = info['element']
            #Find the smallest id which is free
            used_ids = {val['id'] for name, val in species_info.items() if 'id' in val and val['element'] == element}
            possible_ids = range(1, max(used_ids, default=0) + 2)
            info['id'] = min(set(possible_ids) - set(used_ids))
            #Just append the id to the normed name
            info['normed_name'] += f"-{info['id']}"

    return species_info


def get_parameter_data(xmltree, schema_dict, inpgen_ready=True, write_ids=True, extract_econfig=False, logger=None):
    """
    This routine returns an python dictionary produced from the inp.xml
    file, which contains all the parameters needed to setup a new inp.xml from a inpgen
    input file to produce the same input (for parameters that the inpgen can control)

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param inpgen_ready: Bool, return a dict which can be inputed into inpgen while setting atoms
    :param write_ids: Bool, if True the atom ids are added to the atom namelists
    :param logger: logger object for logging warnings, errors

    :returns: dict, which will lead to the same inp.xml (in case if other defaults,
              which can not be controlled by input for inpgen, were changed)

    """
    from masci_tools.util.schema_dict_util import read_constants, eval_simple_xpath, attrib_exists
    from masci_tools.util.schema_dict_util import evaluate_attribute, evaluate_text, evaluate_tag
    from masci_tools.util.xml.common_functions import clear_xml
    from masci_tools.util.xml.converters import convert_fleur_lo, convert_fleur_electronconfig
    from masci_tools.io.common_functions import filter_out_empty_dict_entries

    # TODO: convert econfig
    # TODO: parse kpoints, somehow count is bad (if symmetry changes), mesh is not known, path cannot be specified

    ########
    parameters = {}
    if isinstance(xmltree, etree._ElementTree):
        xmltree, _ = clear_xml(xmltree)
        root = xmltree.getroot()
    else:
        root = xmltree
    constants = read_constants(root, schema_dict, logger=logger)

    # Create the cards

    # &input # most things are not needed for AiiDA here. or we ignor them for now.
    # film is set by the plugin depended on the structure
    # symor per default = False? to avoid input which fleur can't take

    # &comp
    comp_dict = {}
    comp_dict['jspins'] = evaluate_attribute(root, schema_dict, 'jspins', constants=constants, logger=logger)
    comp_dict['frcor'] = evaluate_attribute(root,
                                            schema_dict,
                                            'frcor',
                                            constants=constants,
                                            logger=logger,
                                            optional=True)
    comp_dict['ctail'] = evaluate_attribute(root, schema_dict, 'ctail', constants=constants, logger=logger)
    comp_dict['kcrel'] = evaluate_attribute(root,
                                            schema_dict,
                                            'kcrel',
                                            constants=constants,
                                            logger=logger,
                                            optional=True)
    comp_dict['gmax'] = evaluate_attribute(root, schema_dict, 'gmax', constants=constants, logger=logger)
    comp_dict['gmaxxc'] = evaluate_attribute(root, schema_dict, 'gmaxxc', constants=constants, logger=logger)
    comp_dict['kmax'] = evaluate_attribute(root, schema_dict, 'kmax', constants=constants, logger=logger)
    parameters['comp'] = filter_out_empty_dict_entries(comp_dict)

    # &atoms
    species_list = eval_simple_xpath(root, schema_dict, 'species', list_return=True, logger=logger)

    species_info = _get_species_info(xmltree, schema_dict, logger=logger)

    for indx, species in enumerate(species_list):
        atom_dict = {}
        atoms_name = f'atom{indx}'
        atom_z = evaluate_attribute(species, schema_dict, 'atomicNumber', constants=constants, logger=logger)
        atom_name = evaluate_attribute(species, schema_dict, 'name', constants=constants, logger=logger)
        if not inpgen_ready:
            atom_dict['z'] = atom_z
        atom_id = f"{atom_z}.{species_info[atom_name]['id']}"
        if write_ids:
            atom_dict['id'] = atom_id

        if schema_dict.inp_version <= (0, 31):
            atom_dict['ncst'] = evaluate_attribute(species, schema_dict, 'coreStates', constants, logger=logger)
        atom_dict['rmt'] = evaluate_attribute(species, schema_dict, 'radius', constants=constants, logger=logger)
        atom_dict['dx'] = evaluate_attribute(species, schema_dict, 'logIncrement', constants=constants, logger=logger)
        atom_dict['jri'] = evaluate_attribute(species, schema_dict, 'gridPoints', constants=constants, logger=logger)
        atom_dict['lmax'] = evaluate_attribute(species, schema_dict, 'lmax', constants=constants, logger=logger)
        atom_dict['lnonsph'] = evaluate_attribute(species, schema_dict, 'lnonsphr', constants=constants, logger=logger)
        atom_dict['bmu'] = evaluate_attribute(species, schema_dict, 'magMom', constants, logger=logger, optional=True)

        atom_dict['element'] = evaluate_attribute(species, schema_dict, 'element', constants=constants, logger=logger)

        if extract_econfig:
            if inpgen_ready:
                atom_econfig = eval_simple_xpath(species, schema_dict, 'electronConfig', logger=logger)
                if len(atom_econfig) != 0:
                    atom_dict['econfig'] = convert_fleur_electronconfig(atom_econfig)
            else:
                atom_dict['econfig'] = evaluate_tag(species,
                                                    schema_dict,
                                                    'electronConfig',
                                                    constants=constants,
                                                    logger=logger,
                                                    subtags=True,
                                                    ignore={'flipSpins'})

        atom_lo = eval_simple_xpath(species, schema_dict, 'lo', list_return=True, logger=logger)

        if len(atom_lo) != 0:
            atom_dict['lo'] = convert_fleur_lo(atom_lo)

        parameters[atoms_name] = filter_out_empty_dict_entries(atom_dict)

    # &soc
    soc = evaluate_attribute(root, schema_dict, 'l_soc', constants=constants, logger=logger, optional=True)
    theta = evaluate_attribute(root,
                               schema_dict,
                               'theta',
                               constants=constants,
                               contains='soc',
                               logger=logger,
                               optional=True)
    phi = evaluate_attribute(root,
                             schema_dict,
                             'phi',
                             constants=constants,
                             contains='soc',
                             logger=logger,
                             optional=True)
    if soc is not None and soc:
        parameters['soc'] = {'theta': theta, 'phi': phi}

    # kpt
    if schema_dict.inp_version > (0, 31):
        list_name = evaluate_attribute(root, schema_dict, 'listName', logger=logger)
        kpointlists = eval_simple_xpath(root, schema_dict, 'kPointList', list_return=True, logger=logger)

        if len(kpointlists) == 0:
            raise ValueError('No Kpoint lists found in the given inp.xml')
        labels = [kpoint_set.attrib.get('name') for kpoint_set in kpointlists]
        if list_name not in labels:
            raise ValueError(f'Selected Kpoint list with the name: {list_name} does not exist'
                             f'Available list names: {labels}')

        kpoint_index = labels.index(list_name)
        kpoint_set = kpointlists[kpoint_index]

        if attrib_exists(kpoint_set, schema_dict, 'type', logger=logger):
            kpoint_type = evaluate_attribute(kpoint_set, schema_dict, 'type', logger=logger)

            if kpoint_type == 'mesh':
                nx = evaluate_attribute(kpoint_set, schema_dict, 'nx', logger=logger, optional=True)
                ny = evaluate_attribute(kpoint_set, schema_dict, 'ny', logger=logger, optional=True)
                nz = evaluate_attribute(kpoint_set, schema_dict, 'nz', logger=logger, optional=True)
                if all(n is not None for n in (nx, ny, nz)):
                    parameters['kpt'] = {'div1': nx, 'div2': ny, 'div3': nz}

    # title
    title = evaluate_text(root, schema_dict, 'comment', constants=constants, logger=logger, optional=True)
    if title:
        parameters['title'] = title.replace('\n', '').strip()

    # &exco
    #TODO, easy
    exco_dict = {}
    exco_dict['xctyp'] = evaluate_attribute(root,
                                            schema_dict,
                                            'name',
                                            constants,
                                            contains='xcFunctional',
                                            logger=logger)
    # 'exco' : ['xctyp', 'relxc'],
    parameters['exco'] = filter_out_empty_dict_entries(exco_dict)
    # &film
    # TODO

    # &qss
    # TODO

    # lattice, not supported?

    return parameters


def get_structure_data(xmltree,
                       schema_dict,
                       include_relaxations=True,
                       site_namedtuple=False,
                       convert_to_angstroem=True,
                       normalize_kind_name=True,
                       logger=None):
    """
    Get the structure defined in the given fleur xml file.

    .. warning::
        Only the explicit definition of the Bravais matrix is supported.
        Old inputs containing the `latnam` definitions are not supported

    .. warning::
        In versions ``0.5.0`` or later the output of the atom sites was restructured
        to be more interoperable with other IO functions (e.g. :py:func:`~masci_tools.io.fleur_inpgen.write_inpgen_file()`)
        The new format returns a list of :py:class:`~masci_tools.io.common_functions.AtomSiteProperties`
        instead of the list of tuples (position, symbol)

        For better compatibility this output is not default in ``0.5.0`` but instead
        is enabled by ``site_namedtuple=True`` and a DeprecationWarning is given when
        this argument is ``False``.

    .. note::
        In versions ``0.5.0`` or later the returned atom positions correspond to the relaxed
        structure if a ``relaxation`` section is present in the xmltree


    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param include_relaxations: bool if True and a relaxation section is included
                                the resulting positions correspond to the relaxed structure
    :param logger: logger object for logging warnings, errors
    :param convert_to_angstroem: bool if True the bravais matrix is converted to angstroem

    :returns: tuple containing the structure information

    The tuple contains the following entries:

        1. :atom_data: list of (named)tuples containing the absolute positions and symbols of the atoms
        2. :cell: numpy array, bravais matrix of the given system
        3. :pbc: list of booleans, determines in which directions periodic boundary conditions are applicable

    """
    from masci_tools.util.schema_dict_util import read_constants, eval_simple_xpath, tag_exists
    from masci_tools.util.schema_dict_util import evaluate_text, evaluate_attribute
    from masci_tools.util.xml.common_functions import clear_xml
    from masci_tools.io.common_functions import rel_to_abs, rel_to_abs_f, abs_to_rel, abs_to_rel_f
    from masci_tools.io.common_functions import find_symmetry_relation, AtomSiteProperties
    from masci_tools.util.constants import BOHR_A
    import numpy as np

    if not site_namedtuple:
        warnings.warn(
            'Output of atom positions in pure tuples of the form (position, symbol) is deprecated.'
            'Please adjust your code to use the namedtuple AtomSiteProperties (see masci_tools.io.common_functions)'
            ' with the fields (position, symbol, kind)', DeprecationWarning)

    if isinstance(xmltree, etree._ElementTree):
        xmltree, _ = clear_xml(xmltree)
        root = xmltree.getroot()
    else:
        root = xmltree
    constants = read_constants(root, schema_dict, logger=logger)
    cell, pbc = get_cell(root, schema_dict, logger=logger, convert_to_angstroem=convert_to_angstroem)

    species_info = _get_species_info(xmltree, schema_dict, logger=None)

    atom_data = []
    atom_groups = eval_simple_xpath(root, schema_dict, 'atomGroup', list_return=True, logger=logger)

    #Read relaxation information if available
    displacements = None
    if include_relaxations and schema_dict.inp_version >= (0, 29):
        if tag_exists(root, schema_dict, 'relaxation', logger=logger):
            relax_info = get_relaxation_information(root, schema_dict, logger=logger)
            #We still read in the normal atom positions since the displacements are provided
            #per atomtype
            displacements = relax_info['displacements']
            if convert_to_angstroem:
                displacements = [np.array(displace) * BOHR_A for displace in displacements]
            rotations, shifts = get_symmetry_information(root, schema_dict, logger=logger)

            if len(displacements) != len(atom_groups):
                raise ValueError(
                    f'Did not get the right number of relaxed positions. Expected {len(atom_groups)} got {len(displacements)}'
                )

    for indx, group in enumerate(atom_groups):

        atom_positions = []

        absolute_positions = evaluate_text(group,
                                           schema_dict,
                                           'absPos',
                                           constants=constants,
                                           list_return=True,
                                           logger=logger,
                                           optional=True)
        relative_positions = evaluate_text(group,
                                           schema_dict,
                                           'relPos',
                                           constants=constants,
                                           list_return=True,
                                           logger=logger,
                                           optional=True)
        film_positions = evaluate_text(group,
                                       schema_dict,
                                       'filmPos',
                                       constants=constants,
                                       list_return=True,
                                       logger=logger,
                                       optional=True)

        if convert_to_angstroem:
            atom_positions = [list(np.array(pos) * BOHR_A) for pos in absolute_positions]
        else:
            atom_positions = absolute_positions

        for rel_pos in relative_positions:
            atom_positions.append(rel_to_abs(rel_pos, cell))

        for film_pos in film_positions:
            film_pos = rel_to_abs_f(film_pos, cell)
            if convert_to_angstroem:
                film_pos[2] *= BOHR_A
            atom_positions.append(film_pos)

        if len(atom_positions) == 0:
            raise ValueError('Failed to read atom positions for group')

        if displacements:
            representative_pos = atom_positions[0]

            if len(film_positions) != 0:
                rel_displace = abs_to_rel_f(displacements[indx], cell, pbc)
                rel_representative_pos = abs_to_rel_f(representative_pos, cell, pbc)
                rel_displace[2] = rel_displace[2] / cell[2, 2]
                rel_representative_pos[2] = rel_representative_pos[2] / cell[2, 2]
            else:
                rel_displace = abs_to_rel(displacements[indx], cell)
                rel_representative_pos = abs_to_rel(representative_pos, cell)

            rel_representative_pos = np.array(rel_representative_pos)

            for pos_indx, pos in enumerate(atom_positions):
                rot, shift = find_symmetry_relation(representative_pos,
                                                    pos,
                                                    rotations,
                                                    shifts,
                                                    cell,
                                                    relative_pos=False,
                                                    film=len(film_positions) != 0)

                #More explicit than it needs to be
                #but analogous to fleur
                rot_pos = np.matmul(rot, rel_representative_pos) + shift
                site_displace = np.matmul(rot, rel_representative_pos + rel_displace) + shift
                site_displace = site_displace - rot_pos

                if len(film_positions) != 0:
                    site_displace = rel_to_abs_f(site_displace, cell)
                    site_displace[2] *= cell[2, 2]
                else:
                    site_displace = rel_to_abs(site_displace, cell)

                atom_positions[pos_indx] = list(np.array(atom_positions[pos_indx]) + np.array(site_displace))

        group_species = evaluate_attribute(group, schema_dict, 'species', constants=constants, logger=logger)
        element = species_info[group_species]['element']
        if normalize_kind_name and site_namedtuple:
            normed_name = species_info[group_species]['normed_name']
            if normed_name != group_species:
                if logger is None:
                    warnings.warn(
                        f'Normalized species name {group_species} to {normed_name}. '
                        "Use the option 'normed_kind_name=False' to preserve the original species name", UserWarning)
                else:
                    logger.warning(f'Normalized species name {group_species} to {normed_name}. '
                                   "Use the option 'normed_kind_name=False' to preserve the original species name")
                group_species = normed_name

        if site_namedtuple:
            atom_data.extend(
                AtomSiteProperties(position=pos, symbol=element, kind=group_species) for pos in atom_positions)
        else:
            atom_data.extend((pos, element) for pos in atom_positions)

    return atom_data, cell, pbc


@schema_dict_version_dispatch(output_schema=False)
def get_kpoints_data(xmltree,
                     schema_dict,
                     name=None,
                     index=None,
                     only_used=False,
                     logger=None,
                     convert_to_angstroem=True):
    """
    Get the kpoint sets defined in the given fleur xml file.

    .. warning::
        For file versions before Max5 the name argument is not valid

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param name: str, optional, if given only the kpoint set with the given name
                 is returned
    :param index: int, optional, if given only the kpoint set with the given index
                  is returned
    :param only_used: bool if True only the kpoint list used in the calculation is returned
    :param logger: logger object for logging warnings, errors
    :param convert_to_angstroem: bool if True the bravais matrix is converted to angstroem

    :returns: tuple containing the kpoint information

    The tuple contains the following entries:

        1. :kpoints: dict or list (list if there is only one kpoint set),
                     containing the coordinates of the kpoints
        2. :weights: dict or list (list if there is only one kpoint set),
                     containing the weights of the kpoints
        3. :cell: numpy array, bravais matrix of the given system
        4. :pbc: list of booleans, determines in which directions periodic boundary conditions are applicable

    """
    from masci_tools.util.schema_dict_util import read_constants, eval_simple_xpath
    from masci_tools.util.schema_dict_util import evaluate_text, evaluate_attribute
    from masci_tools.util.xml.common_functions import clear_xml

    if name is not None and index is not None:
        raise ValueError('Only provide one of index or name to select kpoint lists')

    if only_used and (name is not None or index is not None):
        raise ValueError('Either use only_used=False and provide the name/index or use only_used=True. Not both')

    if isinstance(xmltree, etree._ElementTree):
        xmltree, _ = clear_xml(xmltree)
        root = xmltree.getroot()
    else:
        root = xmltree

    constants = read_constants(root, schema_dict, logger=logger)

    if only_used:
        name = evaluate_attribute(root, schema_dict, 'listName', logger=logger)

    cell, pbc = get_cell(root, schema_dict, logger=logger, convert_to_angstroem=convert_to_angstroem)

    kpointlists = eval_simple_xpath(root, schema_dict, 'kPointList', list_return=True, logger=logger)

    if len(kpointlists) == 0:
        raise ValueError('No Kpoint lists found in the given inp.xml')

    labels = [kpoint_set.attrib.get('name') for kpoint_set in kpointlists]
    if name is not None and name not in labels:
        if only_used:
            raise ValueError(f'Found no Kpoint list with the name: {name}'
                             f'Available list names: {labels}'
                             'The listName attribute is not consistent with the rest of the input')
        else:
            raise ValueError(f'Found no Kpoint list with the name: {name}' f'Available list names: {labels}')

    if index is not None:
        try:
            kpointlists = [kpointlists[index]]
        except IndexError as exc:
            raise ValueError(f'No kPointList with index {index} found.' f' Only {len(kpointlists)} available') from exc

    kpoints_data = {}
    weights_data = {}
    for kpointlist in kpointlists:

        label = evaluate_attribute(kpointlist, schema_dict, 'name', logger=logger)

        if name is not None and name != label:
            continue

        kpoints = evaluate_text(kpointlist, schema_dict, 'kPoint', constants=constants, list_return=True, logger=logger)
        weights = evaluate_attribute(kpointlist,
                                     schema_dict,
                                     'weight',
                                     constants=constants,
                                     list_return=True,
                                     logger=logger)

        if not isinstance(kpoints[0], list):
            kpoints = [kpoints]
            weights = [weights]

        kpoints_data[label] = kpoints
        weights_data[label] = weights

    if len(kpoints_data) == 1:
        _, kpoints_data = kpoints_data.popitem()
        _, weights_data = weights_data.popitem()

    return kpoints_data, weights_data, cell, pbc


@get_kpoints_data.register(max_version='0.31')
def get_kpoints_data_max4(xmltree, schema_dict, logger=None, convert_to_angstroem=True, only_used=False):
    """
    Get the kpoint sets defined in the given fleur xml file.

    .. note::
        This function is specific to file version before and including the
        Max4 release of fleur

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors
    :param convert_to_angstroem: bool if True the bravais matrix is converted to angstroem
    :param only_used: (Has no effect for Max4) bool if True only the kpoint list used in the calculation is returned

    :returns: tuple containing the kpoint information

    The tuple contains the following entries:

        1. :kpoints: list containing the coordinates of the kpoints
        2. :weights: list containing the weights of the kpoints
        3. :cell: numpy array, bravais matrix of the given system
        4. :pbc: list of booleans, determines in which directions periodic boundary conditions are applicable

    """
    from masci_tools.util.schema_dict_util import read_constants, eval_simple_xpath
    from masci_tools.util.schema_dict_util import evaluate_text, evaluate_attribute
    from masci_tools.util.xml.common_functions import clear_xml

    if isinstance(xmltree, etree._ElementTree):
        xmltree, _ = clear_xml(xmltree)
        root = xmltree.getroot()
    else:
        root = xmltree

    constants = read_constants(root, schema_dict, logger=logger)

    cell, pbc = get_cell(root, schema_dict, logger=logger, convert_to_angstroem=convert_to_angstroem)

    kpointlist = eval_simple_xpath(root,
                                   schema_dict,
                                   'kPointList',
                                   list_return=True,
                                   not_contains='altKPoint',
                                   logger=logger)

    if len(kpointlist) == 0:
        raise ValueError('No Kpoint lists found in the given inp.xml')

    kpointlist = kpointlist[0]

    kpoints = evaluate_text(kpointlist,
                            schema_dict,
                            'kPoint',
                            constants=constants,
                            not_contains='altKPoint',
                            list_return=True,
                            logger=logger)
    weights = evaluate_attribute(kpointlist,
                                 schema_dict,
                                 'weight',
                                 constants=constants,
                                 not_contains='altKPoint',
                                 list_return=True,
                                 logger=logger)

    return kpoints, weights, cell, pbc


@schema_dict_version_dispatch(output_schema=False)
def get_relaxation_information(xmltree, schema_dict, logger=None):
    """
    Get the relaxation information from the given fleur XML file. This includes the current
    displacements, energy and posforce evolution

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors

    :returns: dict with the relaxation information

    :raises ValueError: If no relaxation section is included in the xml tree
    """
    from masci_tools.util.schema_dict_util import tag_exists, read_constants, evaluate_text, eval_simple_xpath
    from masci_tools.util.schema_dict_util import evaluate_attribute
    from masci_tools.util.xml.common_functions import clear_xml

    if isinstance(xmltree, etree._ElementTree):
        xmltree, _ = clear_xml(xmltree)
        root = xmltree.getroot()
    else:
        root = xmltree
    constants = read_constants(root, schema_dict, logger=logger)

    if not tag_exists(root, schema_dict, 'relaxation', logger=logger):
        raise ValueError('No relaxation information included in the given XML file')

    relax_tag = eval_simple_xpath(root, schema_dict, 'relaxation', logger=logger)

    out_dict = {}
    out_dict['displacements'] = evaluate_text(relax_tag,
                                              schema_dict,
                                              'displace',
                                              list_return=True,
                                              constants=constants,
                                              logger=logger)

    out_dict['energies'] = evaluate_attribute(relax_tag,
                                              schema_dict,
                                              'energy',
                                              list_return=True,
                                              constants=constants,
                                              logger=logger)

    out_dict['posforces'] = []
    relax_iters = eval_simple_xpath(relax_tag, schema_dict, 'step', list_return=True, logger=logger)
    for step in relax_iters:
        posforces = evaluate_text(step, schema_dict, 'posforce', list_return=True, constants=constants, logger=logger)
        out_dict['posforces'].append(posforces)

    return out_dict


@get_relaxation_information.register(max_version='0.28')
def get_relaxation_information_pre029(xmltree, schema_dict, logger=None):
    """
    Get the relaxation information from the given fleur XML file. This includes the current
    displacements, energy and posforce evolution

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors

    :returns: dict with the relaxation information

    :raises ValueError: If no relaxation section is included in the xml tree
    """
    raise NotImplementedError(
        f"'get_relaxation_information' is not implemented for inputs of version '{schema_dict['inp_version']}'")


def get_symmetry_information(xmltree, schema_dict, logger=None):
    """
    Get the symmetry information from the given fleur XML file. This includes the
    rotation matrices and shifts defined in the ``symmetryOperations`` tag.

    .. note::
        Only the explicit definition of the used symmetry operations in the xml file
        is supported.

    :param xmltree: etree representing the fleur xml file
    :param schema_dict: schema dictionary corresponding to the file version
                        of the xmltree
    :param logger: logger object for logging warnings, errors

    :returns: tuple of the rotations and their respective shifts

    :raises ValueError: If no symmetryOperations section is included in the xml tree
    """
    from masci_tools.util.schema_dict_util import tag_exists, read_constants, evaluate_text, eval_simple_xpath
    from masci_tools.util.xml.common_functions import clear_xml
    import numpy as np

    if isinstance(xmltree, etree._ElementTree):
        xmltree, _ = clear_xml(xmltree)
        root = xmltree.getroot()
    else:
        root = xmltree
    constants = read_constants(root, schema_dict, logger=logger)

    if not tag_exists(root, schema_dict, 'symmetryOperations', logger=logger):
        raise ValueError('No explicit symmetry information included in the given XML file')

    ops = eval_simple_xpath(root, schema_dict, 'symOp', logger=logger, list_return=True)

    rotations = []
    shifts = []
    for op in ops:
        row1 = evaluate_text(op, schema_dict, 'row-1', constants=constants, logger=logger)
        row2 = evaluate_text(op, schema_dict, 'row-2', constants=constants, logger=logger)
        row3 = evaluate_text(op, schema_dict, 'row-3', constants=constants, logger=logger)

        rot = np.array([row1[:3], row2[:3], row3[:3]]).astype(int)
        shift = np.array([row1[3], row2[3], row3[3]])

        rotations.append(rot)
        shifts.append(shift)

    return rotations, shifts



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module defines subclasses of UserDict and UserList to be able to prevent
unintended modifications
"""
from collections import UserDict, UserList
from contextlib import contextmanager

from typing import Union, Any, Generator, Iterable, cast


@contextmanager
def LockContainer(lock_object: Union['LockableList', 'LockableDict']) -> Generator:
    """
    Contextmanager for temporarily locking a lockable object. Object is unfrozen
    when exiting with block

    :param lock_object: lockable container (not yet frozen)
    """

    assert isinstance(lock_object, (LockableDict, LockableList)), f'Wrong type Got: {lock_object.__class__}'

    assert not lock_object.locked, f'{lock_object.__class__.__name__} was already locked before entering the contextmanager'

    lock_object.freeze()

    try:
        yield
    finally:
        lock_object._unfreeze()


class LockableDict(UserDict):
    """
    Subclass of UserDict, which can prevent modifications to itself.
    Raises `RuntimeError` if modification is attempted.

    Use :py:meth:`LockableDict.freeze()` to enforce. :py:meth:`LockableDict.get_unlocked()`
    returns a copy of the locked object with builtin lists and dicts

    :param recursive: bool if True (default) all subitems (lists or dicts) are converted into their
                      lockable counterparts

    All other args or kwargs will be passed on to initialize the `UserDict`

    IMPORTANT NOTE:
        This is not a direct subclass of dict. So isinstance(a, dict)
        will be False if a is an LockableDict

    """

    def __init__(self, *args: dict, recursive: bool = True, **kwargs: object) -> None:
        self._locked = False
        self._recursive = recursive
        super().__init__(*args, **kwargs)

    def __check_lock(self) -> None:
        if self.locked:
            raise RuntimeError('Modification not allowed')

    @property
    def locked(self) -> bool:
        """
        Returns whether the object is locked
        """
        return self._locked

    def __delitem__(self, key: object) -> None:
        self.__check_lock()
        super().__delitem__(key)

    def __setitem__(self, key: object, value: object):
        self.__check_lock()
        if isinstance(value, list):
            super().__setitem__(key, LockableList(value, recursive=self._recursive))
        elif isinstance(value, dict):
            super().__setitem__(key, LockableDict(value, recursive=self._recursive))
        else:
            super().__setitem__(key, value)

    def freeze(self) -> None:
        """
        Freezes the object. This prevents further modifications
        """
        if self._recursive:
            for key, val in self.items():
                if isinstance(val, (LockableDict, LockableList)):
                    val.freeze()

        self._locked = True

    def _unfreeze(self) -> None:

        if self._recursive:
            for key, val in self.items():
                if isinstance(val, (LockableList, LockableDict)):
                    val._unfreeze()

        self._locked = False

    def get_unlocked(self) -> dict:
        """
        Get copy of object with builtin lists and dicts
        """
        if self._recursive:
            ret_dict = {}
            for key, value in self.items():
                if isinstance(value, LockableDict):
                    ret_dict[key] = value.get_unlocked()
                elif isinstance(value, LockableList):
                    ret_dict[key] = cast(Any, value.get_unlocked())
                else:
                    ret_dict[key] = value
        else:
            ret_dict = dict(self)

        return ret_dict


class LockableList(UserList):
    """
    Subclass of UserList, which can prevent modifications to itself.
    Raises `RuntimeError` if modification is attempted.

    Use :py:meth:`LockableList.freeze()` to enforce. :py:meth:`LockableList.get_unlocked()`
    returns a copy of the locked object with builtin lists and dicts

    :param recursive: bool if True (default) all subitems (lists or dicts) are converted into their
                      lockable counterparts

    All other args or kwargs will be passed on to initialize the `UserList`

    IMPORTANT NOTE:
        This is not a direct subclass of list. So isinstance(a, list)
        will be False if a is an LockableList

    """

    def __init__(self, *args: Iterable[Any], recursive: bool = True, **kwargs: Iterable[Any]) -> None:
        self._locked = False
        self._recursive = recursive
        super().__init__(*args, **kwargs)
        if self._recursive:
            #Convert sublists and subdicts into Lockable counterparts (super__init__ just copies the values)
            for indx, item in enumerate(self):
                if isinstance(item, list):
                    super().__setitem__(indx, LockableList(item, recursive=self._recursive))
                elif isinstance(item, dict):
                    super().__setitem__(indx, LockableDict(item, recursive=self._recursive))

    def __check_lock(self) -> None:
        if self.locked:
            raise RuntimeError('Modification not allowed')

    @property
    def locked(self) -> bool:
        """
        Returns whether the object is locked
        """
        return self._locked

    def __delitem__(self, i: Union[int, slice]) -> None:
        self.__check_lock()
        super().__delitem__(i)

    def __setitem__(self, i: Union[int, slice], item: object) -> None:
        self.__check_lock()
        if isinstance(item, list):
            super().__setitem__(i, LockableList(item, recursive=self._recursive))
        elif isinstance(item, dict):
            super().__setitem__(i, LockableDict(item, recursive=self._recursive))
        else:
            super().__setitem__(i, item)  # type: ignore

    def __iadd__(self, other: Iterable[Any]) -> 'LockableList':
        self.__check_lock()
        return super().__iadd__(other)

    def __imul__(self, n: int) -> 'LockableList':
        self.__check_lock()
        return super().__imul__(n)

    def append(self, item: object) -> None:
        self.__check_lock()
        super().append(item)

    def insert(self, i: int, item: object) -> None:
        self.__check_lock()
        super().insert(i, item)

    def pop(self, i: int = -1) -> object:
        """
        return the value at index i (default last) and remove it from list
        """
        self.__check_lock()
        return super().pop(i=i)

    def remove(self, item: object) -> None:
        self.__check_lock()
        super().remove(item)

    def clear(self) -> None:
        """
        Clear the list
        """
        self.__check_lock()
        super().clear()

    def reverse(self) -> None:
        self.__check_lock()
        super().reverse()

    def sort(self, *args: object, **kwargs: object) -> None:
        self.__check_lock()
        super().sort(*args, **kwargs)

    def extend(self, other: Iterable[object]) -> None:
        self.__check_lock()
        super().extend(other)

    def freeze(self) -> None:
        """
        Freezes the object. This prevents further modifications
        """
        if self._recursive:
            for val in self:
                if isinstance(val, (LockableList, LockableDict)):
                    val.freeze()

        self._locked = True

    def _unfreeze(self) -> None:

        if self._recursive:
            for val in self:
                if isinstance(val, (LockableList, LockableDict)):
                    val._unfreeze()

        self._locked = False

    def get_unlocked(self) -> Iterable[Any]:
        """
        Get copy of object with builtin lists and dicts
        """
        if self._recursive:
            ret_list = []
            for value in self:
                if isinstance(value, LockableDict):
                    ret_list.append(cast(Any, value.get_unlocked()))
                elif isinstance(value, LockableList):
                    ret_list.append(value.get_unlocked())
                else:
                    ret_list.append(value)
        else:
            ret_list = list(self)

        return ret_list
